""" Static file server web module for x/84 bbs. """

import web
import os
import mimetypes

class StaticApp(object):

    """ Class for serving static files. """

    static_root = ''
    mime_types = {
        '.html': 'text/html',
        '.css': 'text/css',
        '.js': 'text/javascript',
    }

    def GET(self, filename):
        """ Respond to GET method request. """
        if not filename:
            return web.redirect('/www-static/')
        file_url = os.path.join(*filter(lambda txt: txt != '..',
                                        filename.split('/')))
        myfile = os.path.join(StaticApp.static_root, file_url)
        if os.path.isfile(myfile):
            # we're serving a file; use the proper mime type
            mime = mimetypes.guess_type(myfile)
            _, ext = os.path.splitext(myfile.lower())
            mime = StaticApp.mime_types.get(ext, mime)
            web.header('Content-Type', mime, unique=True)
            return open(myfile, 'rb').read()
        elif os.path.isdir(myfile):
            # we're serving a directory; try directory/index.html instead
            if not filename.endswith('/'):
                return web.redirect(''.join([filename, '/']))
            return self.GET('/'.join([filename, 'index.html']))
        # path does not exist; return 404
        return web.notfound()


def web_module():
    """ Expose our REST API. Run only once on server startup. """
    from x84.bbs.ini import get_ini

    # determine document root for web server
    static_root = (get_ini('web', 'document_root')
                   or os.path.join(get_ini('system', 'scriptpath'),
                                   'www-static'))
    StaticApp.static_root = static_root

    return {
        'urls': ('/www-static(/.*)?', 'static'),
        'funcs': {
            'static': StaticApp
        }
    }



""" Package provides base server for x/84. """


class BaseServer(object):

    """ Base class for server implementations. """

    #: Maximum number of clients
    MAX_CONNECTIONS = 100

    #: Number of clients that can wait to be accepted
    LISTEN_BACKLOG = 5

    #: Dictionary of environment variables received by negotiation
    env = {}

    #: Client factory should be a class defining what should be instantiated
    #: for the client instance.
    client_factory = None

    #: Dictionary of active clients, (file descriptor, Client, ...)
    clients = {}

    #: Connect factory should be a class, derived from threading.Thread, that
    #: should be instantiated on-connect to perform negotiation and launch the
    #: bbs session upon success.
    connect_factory = None

    #: List of on-connect negotiating threads.
    threads = []

    @classmethod
    def client_factory_kwargs(cls, instance):
        """
        Return keyword arguments for the client_factory.

        Method should be derived and modified, A dictionary may be substituted.
        The default return value is an empty dictionary.

        :rtype dict
        """
        # pylint: disable=W0613
        #         Unused argument 'instance'
        return dict()

    @classmethod
    def connect_factory_kwargs(cls, instance):
        """
        Return keyword arguments for the connect_factory.

        Method should be derived and modified, A dictionary may be substituted.
        The default return value is an empty dictionary.

        :rtype dict
        """
        # pylint: disable=W0613
        #         Unused argument 'instance'
        return dict()

    def client_count(self):
        """ Return number of active connections.  """
        return len(self.clients)

    def client_list(self):
        """ Return list of connected clients. """
        return self.clients.values()

    def client_fds(self):
        """ Return list of client file descriptors.  """
        return [client.fileno() for client in self.clients.values()]

    def clients_ready(self, ready_fds=None):
        """
        Return list of clients with data ready to be receive.

        :param list ready_fds: file descriptors already known to be ready
        """
        if ready_fds is None:
            # given no file descriptors, we must iterate them all by hand.
            return [client for client in self.clients.values()
                    if client.recv_ready()]

        # given a list of ready_fds pairs, we return only clients with
        # matching file descriptors.
        return [client for client in self.clients.values()
                if client.fileno() in ready_fds]



""" Database proxy helper for x/84. """
# std imports
import logging

# local
from x84.bbs.ini import get_ini
from x84.db import (
    get_db_filepath,
    get_database,
    get_db_func,
    get_db_lock,
    log_db_cmd,
)


class DBProxy(object):

    """
    Provide dictionary-like object interface to shared database.

    A database call, such as __len__() or keys() is issued as a command
    to the main engine when ``use_session`` is True, which spawns a thread
    to acquire a lock on the database and return the results via IPC pipe
    transfer.
    """

    def __init__(self, schema, table='unnamed', use_session=True):
        """
        Class initializer.

        :param str scheme: database key, becomes basename of .sqlite3 file.
        :param str table: optional database table.
        :param bool use_session: Whether iterable returns should be sent over
                                 an IPC pipe (client is a
                                 :class:`x84.bbs.session.Session` instance),
                                 or returned directly (such as used by the main
                                 thread engine components.)
        """
        self.log = logging.getLogger(__name__)
        self.schema = schema
        self.table = table
        self._tap_db = get_ini('session', 'tab_db', getter='getboolean')

        from x84.bbs.session import getsession
        self._session = use_session and getsession()

    def proxy_iter_session(self, method, *args):
        """ Proxy for iterable-return method calls over session IPC pipe. """
        event = 'db={0}'.format(self.schema)
        self._session.flush_event(event)
        self._session.send_event(event, (self.table, method, args))
        data = self._session.read_event(event)
        assert data == (None, 'StartIteration'), (
            'iterable proxy used on non-iterable, {0!r}'.format(data))
        data = self._session.read_event(event)
        while data != (None, StopIteration):
            yield data
            data = self._session.read_event(event)
        self._session.flush_event(event)

    def proxy_method_direct(self, method, *args):
        """ Proxy for direct dictionary method calls. """
        dictdb = get_database(filepath=get_db_filepath(self.schema),
                              table=self.table)
        try:
            func = get_db_func(dictdb, method)
            if self._tap_db:
                log_db_cmd(self.log, self.schema, method, args)
            return func(*args)
        finally:
            dictdb.close()

    def proxy_iter(self, method, *args):
        """ Proxy for iterable dictionary method calls. """
        if self._session:
            return self.proxy_iter_session(method, *args)

        return self.proxy_method_direct(method, *args)

    def proxy_method(self, method, *args):
        """ Proxy for dictionary method calls. """
        if self._session:
            return self.proxy_method_session(method, *args)

        return self.proxy_method_direct(method, *args)

    def proxy_method_session(self, method, *args):
        """ Proxy for dictionary method calls over IPC pipe. """
        event = 'db-{0}'.format(self.schema)
        self._session.send_event(event, (self.table, method, args))
        return self._session.read_event(event)

    def acquire(self):
        """ Acquire system-wide lock on database. """
        lock = get_db_lock(schema=self.schema, table=self.table)
        if self._tap_db:
            self.log.debug('lock acquire schema=%s, table=%s',
                           self.schema, self.table)
        lock.acquire()

    def release(self):
        """ Release system-wide lock on database. """
        lock = get_db_lock(schema=self.schema, table=self.table)
        if self._tap_db:
            self.log.debug('lock release schema=%s, table=%s',
                           self.schema, self.table)
        lock.release()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.release()

    # pylint: disable=C0111
    #        Missing docstring
    def __contains__(self, key):
        return self.proxy_method('__contains__', key)
    __contains__.__doc__ = dict.__contains__.__doc__

    def __getitem__(self, key):
        return self.proxy_method('__getitem__', key)
    __getitem__.__doc__ = dict.__getitem__.__doc__

    def __setitem__(self, key, value):
        return self.proxy_method('__setitem__', key, value)
    __setitem__.__doc__ = dict.__setitem__.__doc__

    def __delitem__(self, key):
        return self.proxy_method('__delitem__', key)
    __delitem__.__doc__ = dict.__delitem__.__doc__

    def get(self, key, default=None):
        return self.proxy_method('get', key, default)
    get.__doc__ = dict.get.__doc__

    def has_key(self, key):
        return self.proxy_method('has_key', key)
    has_key.__doc__ = dict.has_key.__doc__

    def setdefault(self, key, value):
        return self.proxy_method('setdefault', key, value)
    setdefault.__doc__ = dict.setdefault.__doc__

    def update(self, *args):
        return self.proxy_method('update', *args)
    update.__doc__ = dict.update.__doc__

    def __len__(self):
        return self.proxy_method('__len__')
    __len__.__doc__ = dict.__len__.__doc__

    def values(self):
        return self.proxy_method('values')
    values.__doc__ = dict.values.__doc__

    def items(self):
        return self.proxy_method('items')
    items.__doc__ = dict.items.__doc__

    def iteritems(self):
        return self.proxy_iter('iteritems')
    iteritems.__doc__ = dict.iteritems.__doc__

    def iterkeys(self):
        return self.proxy_iter('iterkeys')
    iterkeys.__doc__ = dict.iterkeys.__doc__

    def itervalues(self):
        return self.proxy_iter('itervalues')
    itervalues.__doc__ = dict.itervalues.__doc__

    def keys(self):
        return self.proxy_method('keys')
    keys.__doc__ = dict.keys.__doc__

    def pop(self):
        return self.proxy_method('pop')
    pop.__doc__ = dict.pop.__doc__

    def popitem(self):
        return self.proxy_method('popitem')
    popitem.__doc__ = dict.popitem.__doc__

    def copy(self):
        # https://github.com/piskvorky/sqlitedict/issues/20
        # @jquast: should sqlitedict have a .copy() method? "no."
        return dict(self.proxy_method('items'))
    copy.__doc__ = dict.copy.__doc__



""" Messaging database package for x/84. """
# std imports
import datetime
import logging

# local
from x84.bbs.dbproxy import DBProxy
from x84.bbs.session import getsession
from x84.bbs.ini import get_ini

# 3rd party
import dateutil.tz

MSGDB = 'msgbase'
TAGDB = 'tags'
PRIVDB = 'privmsg'

# TODO(jquast, maze): Use modeling to construct rfc-compliant mail messaging
# formats.  It would be possible to use standard mbox-formatted mail boxes,
# and integrate with external systems.  This is a v3.0 release.


def to_localtime(tm_value):
    """ convert given UTC time to local time """
    utcz = dateutil.tz.tzutc()
    locz = dateutil.tz.tzlocal()
    utime = datetime.datetime.strptime(tm_value, '%Y-%m-%d %H:%M:%S')
    utime = utime.replace(tzinfo=utcz)
    ltime = utime.astimezone(locz)
    return ltime.replace(tzinfo=None)


def to_utctime(tm_value):
    """ convert given local time to UTC time """
    utcz = dateutil.tz.tzutc()
    locz = dateutil.tz.tzlocal()
    ltime = tm_value.replace(tzinfo=locz)
    utime = ltime.astimezone(utcz)
    return utime.replace(tzinfo=None).isoformat(' ').partition('.')[0]


def get_origin_line():
    """ Return ``origin`` configuration item of ``[msg]`` section. """
    return get_ini(section='msg', key='origin_line') or (
        'Sent from {0}'.format(
            get_ini(section='system', key='bbsname')))


def format_origin_line():
    """ Format origin line for message quoting. """
    return u''.join((u'\r\n---\r\n', get_origin_line()))


def get_msg(idx=0):
    """ Return Msg record instance by index ``idx``. """
    return DBProxy(MSGDB)['%d' % int(idx)]


def list_msgs(tags=None):
    """ Return set of indices matching ``tags``, or all by default. """
    if tags is not None and 0 != len(tags):
        msgs = set()
        db_tag = DBProxy(TAGDB)
        for tag in (_tag for _tag in tags if _tag in db_tag):
            msgs.update(db_tag[tag])
        return msgs
    return set(int(key) for key in DBProxy(MSGDB).keys())


def list_privmsgs(handle=None):
    """ Return all private messages for given user handle. """
    db_priv = DBProxy(PRIVDB)
    if handle:
        return db_priv.get(handle, set())
    # flatten list of [set(1, 2), set(3, 4)] to set(1, 2, 3, 4)
    return set([_idx for indices in db_priv.values() for _idx in indices])


def list_tags():
    """ Return set of available tags. """
    return [_tag.decode('utf8') for _tag in DBProxy(TAGDB).keys()]


class Msg(object):

    """
    A record spec for messages held in the msgbase.

    It contains many default properties to describe a conversation:

    - ``stime``, the time the message was sent.

    - ``author``, ``recipient``, ``subject``, and ``body`` are envelope
      parameters.

    - ``tags`` is for use with message groupings, containing a list of strings
      that other messages may share in relation.

    - ``parent`` points to the message this message directly refers to.

    - ``children`` is a set of indices replied by this message.
    """

    # pylint: disable=R0902
    #         Too many instance attributes
    @property
    def ctime(self):
        """
        Datetime message was instantiated

        :rtype: datetime.datetime
        """
        return self._ctime

    @property
    def stime(self):
        """
        Datetime message was saved to database

        :rtype: datetime.datetime
        """
        return self._stime

    def __init__(self, recipient=None, subject=u'', body=u''):
        self.author = None
        session = getsession()
        if session:
            self.author = session.user.handle

        self._ctime = datetime.datetime.now()
        self._stime = None
        self.recipient = recipient
        self.subject = subject
        self.body = body
        self.tags = set()
        self.children = set()
        self.parent = None
        self.idx = None

    def save(self, send_net=True, ctime=None):
        """
        Save message to database, recording 'tags' db.

        As a side-effect, it may queue message for delivery to
        external systems, when configured.
        """
        log = logging.getLogger(__name__)
        session = getsession()
        use_session = bool(session is not None)
        new = self.idx is None or self._stime is None

        # persist message record to MSGDB
        with DBProxy(MSGDB, use_session=use_session) as db_msg:
            if new:
                self.idx = max(map(int, db_msg.keys()) or [-1]) + 1
                if ctime is not None:
                    self._ctime = self._stime = ctime
                else:
                    self._stime = datetime.datetime.now()
                new = True
            db_msg['%d' % (self.idx,)] = self

        # persist message idx to TAGDB
        with DBProxy(TAGDB, use_session=use_session) as db_tag:
            for tag in db_tag.keys():
                msgs = db_tag[tag]
                if tag in self.tags and self.idx not in msgs:
                    msgs.add(self.idx)
                    db_tag[tag] = msgs
                    log.debug("msg {self.idx} tagged '{tag}'"
                              .format(self=self, tag=tag))
                elif tag not in self.tags and self.idx in msgs:
                    msgs.remove(self.idx)
                    db_tag[tag] = msgs
                    log.info("msg {self.idx} removed tag '{tag}'"
                             .format(self=self, tag=tag))
            for tag in [_tag for _tag in self.tags if _tag not in db_tag]:
                db_tag[tag] = set([self.idx])

        # persist message as child to parent;
        assert self.parent not in self.children, ('circular reference',
                                                  self.parent, self.children)
        if self.parent is not None:
            try:
                parent_msg = get_msg(self.parent)
            except KeyError:
                log.warn('Child message {0}.parent = {1}: '
                         'parent does not exist!'.format(self.idx, self.parent))
            else:
                if self.idx != parent_msg.idx:
                    parent_msg.children.add(self.idx)
                    parent_msg.save()
                else:
                    log.error('Parent idx same as message idx; stripping')
                    self.parent = None
                    with db_msg:
                        db_msg['%d' % (self.idx)] = self

        # persist message record to PRIVDB
        if 'public' not in self.tags:
            with DBProxy(PRIVDB, use_session=use_session) as db_priv:
                db_priv[self.recipient] = (
                    db_priv.get(self.recipient, set()) | set([self.idx]))

        # if either any of 'server_tags' or 'network_tags' are enabled,
        # then queue for potential delivery.
        if send_net and new and (
            get_ini(section='msg', key='network_tags') or
            get_ini(section='msg', key='server_tags')
        ):
            self.queue_for_network()

        log.info(
            u"saved {new} {public_or_private} {message_or_reply}"
            u", addressed to '{self.recipient}'."
            .format(new='new ' if new else '',
                    public_or_private=('public' if 'public' in self.tags
                                       else 'private'),
                    message_or_reply=('message' if self.parent is None
                                      else 'reply'),
                    self=self))

    def queue_for_network(self):
        """ Queue message for networks, hosting or sending. """
        log = logging.getLogger(__name__)

        # check all tags of message; if they match a message network,
        # either record for hosting servers, or schedule for delivery.
        for tag in self.tags:

            # server networks offered by this server,
            # message is for a network we host
            if tag in get_ini(section='msg', key='server_tags', split=True):
                with DBProxy('{0}trans'.format(tag)) as transdb:
                    self.body = u''.join((self.body, format_origin_line()))
                    self.save()
                    transdb[self.idx] = self.idx
                log.info('[{tag}] Stored for network (msgid {self.idx}).'
                         .format(tag=tag, self=self))

            # server networks this server is a member of,
            # message is for a another network, queue for delivery
            elif tag in get_ini(section='msg', key='network_tags', split=True):
                with DBProxy('{0}queues'.format(tag)) as queuedb:
                    queuedb[self.idx] = tag
                log.info('[{tag}] Message (msgid {self.idx}) queued '
                         'for delivery'.format(tag=tag, self=self))



""" Command-line parser for x/84. """
import getopt
import sys
import os


def parse_args():
    """ Parse system arguments and return lookup path for bbs and log ini. """
    if sys.platform.lower().startswith('win32'):
        system_path = os.path.join('C:', 'x84')
    else:
        system_path = os.path.join(os.path.sep, 'etc', 'x84')

    lookup_bbs = (os.path.join(system_path, 'default.ini'),
                  os.path.expanduser(os.path.join('~', '.x84', 'default.ini')))

    lookup_log = (os.path.join(system_path, 'logging.ini'),
                  os.path.expanduser(os.path.join('~', '.x84', 'logging.ini')))

    try:
        opts, tail = getopt.getopt(sys.argv[1:], u'', (
            'config=', 'logger=', 'help'))
    except getopt.GetoptError as err:
        sys.stderr.write('{0}\n'.format(err))
        return 1
    for opt, arg in opts:
        if opt in ('--config',):
            lookup_bbs = (arg,)
        elif opt in ('--logger',):
            lookup_log = (arg,)
        elif opt in ('--help',):
            sys.stderr.write(
                'Usage: \n'
                '{0} [--config <filepath>] [--logger <filepath>]\n'
                .format(os.path.basename(sys.argv[0])))
            sys.exit(1)
    if len(tail):
        sys.stderr.write('Unrecognized program arguments: {0}\n'
                         .format(tail))
        sys.exit(1)
    return (lookup_bbs, lookup_log)






#!/usr/bin/env python
""" Command-line launcher and event loop for x/84. """
# Place ALL metadata in setup.py, except where not suitable, place here.
# For any contributions, feel free to tag __author__ etc. at top of such file.
__author__ = "Johannes Lundberg (jojo), Jeff Quast (dingo)"
__url__ = u'https://github.com/jquast/x84/'
__copyright__ = "Copyright 2003"
__credits__ = [
    # use 'scene' names unless preferred or unavailable.
    "zipe",
    "jojo",
    "maze",
    "dingo",
    "spidy",
    "beardy",
    "haliphax",
    "megagumbo",
    "hellbeard",
    "Mercyful Fate",
]
__license__ = 'ISC'

# std
import logging
import select
import socket
import time
import sys

# local
__import__('encodings')  # provides alternate encodings
from x84 import cmdline
from x84.db import DBHandler
from x84.terminal import get_terminals, kill_session, find_tty
from x84.fail2ban import get_fail2ban_function


def main():
    """
    x84 main entry point. The system begins and ends here.

    Command line arguments to engine.py:

    - ``--config=`` location of alternate configuration file
    - ``--logger=`` location of alternate logging.ini file
    """
    # load existing .ini files or create default ones.
    import x84.bbs.ini
    x84.bbs.ini.init(*cmdline.parse_args())

    from x84.bbs import get_ini
    from x84.bbs.ini import CFG

    if sys.maxunicode == 65535:
        # apple is the only known bastardized variant that does this;
        # presumably for memory/speed savings (UCS-2 strings are faster
        # than UCS-4).  Python 3 dynamically allocates string types by
        # their widest content, so such things aren't necessary, there.
        import warnings
        warnings.warn('This python is built without wide unicode support. '
                      'some internationalized languages will not be possible.')

    # retrieve list of managed servers
    servers = get_servers(CFG)

    # begin unmanaged servers
    if (CFG.has_section('web') and
            (not CFG.has_option('web', 'enabled')
             or CFG.getboolean('web', 'enabled'))):
        # start https server for one or more web modules.
        from x84 import webserve
        webserve.main()

    if get_ini(section='msg', key='network_tags'):
        # start background timer to poll for new messages
        # of message networks we may be a member of.
        from x84 import msgpoll
        msgpoll.main()

    try:
        # begin main event loop
        _loop(servers)
    except KeyboardInterrupt:
        # exit on ^C, killing any client sessions.
        for server in servers:
            for thread in server.threads[:]:
                if not thread.stopped:
                    thread.stopped = True
                server.threads.remove(thread)
            for key, client in server.clients.items()[:]:
                kill_session(client, 'server shutdown')
                del server.clients[key]
    return 0


def get_servers(CFG):
    """ Instantiate and return enabled servers by configuration ``CFG``. """
    servers = []

    if (CFG.has_section('telnet') and
            (not CFG.has_option('telnet', 'enabled')
             or CFG.getboolean('telnet', 'enabled'))):
        # start telnet server instance
        from x84.telnet import TelnetServer
        servers.append(TelnetServer(config=CFG))

    if (CFG.has_section('ssh') and
            not CFG.has_option('ssh', 'enabled')
            or CFG.getboolean('ssh', 'enabled')):
        # start ssh server instance
        #
        # may raise an ImportError for systems where pyOpenSSL and etc. could
        # not be installed (due to any issues with missing python-dev, libffi,
        # cc, etc.).  Allow it to raise naturally, the curious user should
        # either discover and resolve the root issue, or disable ssh if it
        # cannot be resolved.
        from x84.ssh import SshServer
        servers.append(SshServer(config=CFG))

    if (CFG.has_section('rlogin') and
            (not CFG.has_option('rlogin', 'enabled')
             or CFG.getboolean('rlogin', 'enabled'))):
        # start rlogin server instance
        from x84.rlogin import RLoginServer
        servers.append(RLoginServer(config=CFG))

    return servers


def find_server(servers, fd):
    """ Find matching ``server.server_socket`` for given file descriptor. """
    for server in servers:
        if fd == server.server_socket.fileno():
            return server


def accept(log, server, check_ban):
    """
    Accept new connection from server, spawning an unmanaged thread.

    Connecting socket accepted is server.server_socket, instantiate a
    new instance of client_factory, with optional keyword arguments
    defined by server.client_factory_kwargs, registering it with
    dictionary server.clients, and spawning an unmanaged thread
    using connect_factory, with optional keyword arguments
    server.connect_factory_kwargs.
    """
    if None in (server.client_factory, server.connect_factory):
        raise NotImplementedError(
            "No accept for server class {server.__class__.__name__}"
            .format(server=server))

    client_factory_kwargs = server.client_factory_kwargs
    if callable(server.client_factory_kwargs):
        client_factory_kwargs = server.client_factory_kwargs(server)

    connect_factory_kwargs = server.connect_factory_kwargs
    if callable(server.connect_factory_kwargs):
        connect_factory_kwargs = server.connect_factory_kwargs(server)

    try:
        sock, address_pair = server.server_socket.accept()

        # busy signal
        if server.client_count() > server.MAX_CONNECTIONS:
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except socket.error:
                pass
            sock.close()
            log.error('{addr}: refused, maximum connections reached.'
                      .format(addr=address_pair[0]))
            return

        # connecting IP is banned
        if check_ban(address_pair[0]) is False:
            log.debug('{addr}: refused, banned.'.format(addr=address_pair[0]))
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except socket.error:
                pass
            sock.close()
            return

        # instantiate a client of this type
        client = server.client_factory(sock, address_pair,
                                       **client_factory_kwargs)

        # spawn on-connect negotiation thread.  When successful,
        # a new sub-process is spawned and registered as a session tty.
        server.clients[client.sock.fileno()] = client
        thread = server.connect_factory(client, **connect_factory_kwargs)
        log.info('{client.kind} connection from {client.addrport} '
                 '(*{thread.name}).'.format(client=client, thread=thread))
        server.threads.append(thread)
        thread.start()

    except socket.error as err:
        log.error('accept error {0}:{1}'.format(*err))


def get_session_output_fds(servers):
    """ Return file descriptors of all ``tty.master_read`` pipes. """
    session_fds = []
    for server in servers:
        for client in server.clients.values():
            tty = find_tty(client)
            if tty is not None:
                session_fds.append(tty.master_read.fileno())
    return session_fds


def client_recv(servers, ready_fds, log):
    """
    Test all clients for recv_ready().

    If any data is available, then ``client.socket_recv()`` is called,
    buffering the data for the session which is exhausted by
    :func:`session_send`.
    """
    from x84.bbs.exception import Disconnected
    for server in servers:
        for client in server.clients_ready(ready_fds):
            try:
                client.socket_recv()
            except Disconnected as err:
                log.debug('{client.addrport}: disconnect on recv: {err}'
                          .format(client=client, err=err))
                kill_session(client, 'disconnected: {err}'.format(err=err))


def client_send(terminals, log):
    """
    Test all clients for send_ready().

    If any data is available, then ``tty.client.send()`` is called.
    This is data sent from the session to the tcp client.
    """
    from x84.bbs.exception import Disconnected
    # nothing to send until tty is registered.
    for _, tty in terminals:
        if tty.client.send_ready():
            try:
                tty.client.send()
            except Disconnected as err:
                log.debug('{client.addrport}: disconnect on send: {err}'
                          .format(client=tty.client, err=err))
                kill_session(tty.client, 'disconnected: {err}'.format(err=err))


def session_send(terminals):
    """
    Test all tty clients for input_ready().

    Meaning, tcp data has been buffered to be received by the tty session,
    and send it to the tty input queue (tty.master_write).  Also, test all
    sessions for idle timeout, signaling exit to subprocess when reached.
    """
    for _, tty in terminals:
        if tty.client.input_ready():
            try:
                tty.master_write.send(('input', tty.client.get_input()))
            except IOError:
                # this may happen if a sub-process crashes, or more often,
                # because the subprocess has logged off, but the user kept
                # banging the keyboard before we have had the opportunity
                # to close their telnet socket.
                kill_session(tty.client, 'no tty for socket data')

        # poll about and kick off idle users
        elif tty.timeout and tty.client.idle() > tty.timeout:
            kill_session(tty.client, 'timeout')


def handle_lock(locks, tty, event, data, tap_events, log):
    """ handle locking event of ``(lock-key, (method, stale))``. """
    # pylint: disable=R0913
    #         Too many arguments (6/5)
    method, stale = data
    if method == 'acquire':
        # this lock is already held,
        if event in locks:
            # check if lock held by an active session,
            holder = locks[event][1]
            for _sid, _ in get_terminals():
                if _sid == holder and _sid != tty.sid:
                    # acquire the lock from a now-deceased session.
                    log.debug('[{tty.sid}] {event} not acquired, '
                              'held by active session: {holder}'
                              .format(tty=tty, event=event, holder=holder))
                    break
                elif _sid == holder and _sid == tty.sid:
                    # acquire the lock from ourselves!  We'll allow it
                    # (this is termed, "re-entrant locking").
                    log.debug('{tty.sid}] {event} is re-acquired!'
                              .format(tty=tty, event=event))
            else:
                # lock is held by a now-defunct session, re-acquired.
                log.debug('[{tty.sid}] {event} re-acquiring stale lock, '
                          'previously held by session no longer active: '
                          '{holder}'
                          .format(tty=tty, event=event, holder=holder))
                del locks[event]

        # lock is not held, or release by previous block
        if event not in locks:
            # acknowledge its requirement,
            locks[event] = (time.time(), tty.sid)
            tty.master_write.send((event, True,))
            if tap_events:
                log.debug('[{tty.sid}] {event} granted lock.'
                          .format(tty=tty, event=event))

        # lock cannot be acquired,
        else:
            holder = locks[event][1]
            elapsed = time.time() - locks[event][0]
            if (stale is not None and elapsed > stale):
                # caller has decreed that this lock may be acquired even if
                # it already held, if it has been held longer than length of
                # time `stale`.  This is simply to prevent a global freeze
                # when the programmer knows the holder may fail to release,
                # though this is not currently used in the demonstration
                # system.
                locks[event] = (time.time(), tty.sid)
                log.warn('[{tty.sid}] {event} re-acquiring stale lock, '
                         'previously held active session {holder} after '
                         '{elapsed}s elapsed (stale={stale})'
                         .format(tty=tty, event=event, holder=holder,
                                 elapsed=elapsed, stale=stale))
                tty.master_write.send((event, True,))

            # signal busy with matching event, data=False
            else:
                log.debug('[{tty.sid}] {event} lock rejected; already held '
                          'by active session {holder} for {elapsed} seconds '
                          '(stale={stale})'
                          .format(tty=tty, event=event, holder=holder,
                                  elapsed=elapsed, stale=stale))
                tty.master_write.send((event, False,))

    elif method == 'release':
        if event not in locks:
            log.error('[{tty.sid}] {event} lock failed to release, '
                      'not acquired.'.format(tty=tty, event=event))
        else:
            del locks[event]
            if tap_events:
                log.debug('[{tty.sid}] {event} released lock.'
                          .format(tty=tty, event=event))


def session_recv(locks, terminals, log, tap_events):
    """
    Receive data waiting for terminal sessions.

    All data received from subprocess is handled here.
    """
    for sid, tty in terminals:
        while tty.master_read.poll():
            try:
                event, data = tty.master_read.recv()
            except (EOFError, IOError) as err:
                # sub-process unexpectedly closed
                log.exception('master_read pipe: {0}'.format(err))
                kill_session(tty.client, 'master_read pipe: {0}'.format(err))
                break
            except TypeError as err:
                log.exception('unpickling error: {0}'.format(err))
                break

            # 'exit' event, unregisters client
            if event == 'exit':
                kill_session(tty.client, 'client exit')
                break

            # 'logger' event, prefix log message with handle and IP address
            elif event == 'logger':
                data.msg = ('{data.handle}[{tty.sid}] {data.msg}'
                            .format(data=data, tty=tty))
                log.handle(data)

            # 'output' event, buffer for tcp socket
            elif event == 'output':
                tty.client.send_unicode(ucs=data[0], encoding=data[1])

            # 'remote-disconnect' event, hunt and destroy
            elif event == 'remote-disconnect':
                for _sid, _tty in terminals:
                    # data[0] is 'send-to' address.
                    if data[0] == _sid:
                        kill_session(
                            tty.client, 'remote-disconnect by {0}'.format(sid))
                        break

            # 'route': message passing directly from one session to another
            elif event == 'route':
                if tap_events:
                    log.debug('route {0!r}'.format(data))
                tgt_sid, send_event, send_val = data[0], data[1], data[2:]
                for _sid, _tty in terminals:
                    if tgt_sid == _sid:
                        _tty.master_write.send((send_event, send_val))
                        break

            # 'global': message broadcasting to all sessions
            elif event == 'global':
                if tap_events:
                    log.debug('broadcast: {data!r}'.format(data=data))
                for _sid, _tty in terminals:
                    if sid != _sid:
                        _tty.master_write.send((event, data,))

            # 'set-timeout': set user-preferred timeout
            elif event == 'set-timeout':
                if tap_events:
                    log.debug('[{tty.sid}] set-timeout {data}'
                              .format(tty=tty, data=data))
                tty.timeout = data

            # 'db*': access DBProxy API for shared sqlitedict
            elif event.startswith('db'):
                DBHandler(tty.master_write, event, data).start()

            # 'lock': access fine-grained bbs-global locking
            elif event.startswith('lock'):
                handle_lock(locks, tty, event, data, tap_events, log)

            else:
                log.error('[{tty.sid}] unhandled event, data: '
                          '({event}, {data})'
                          .format(tty=tty, event=event, data=data))


def _loop(servers):
    """ Main event loop. Never returns. """
    # pylint: disable=R0912,R0914,R0915
    #         Too many local variables (24/15)
    from x84.bbs.ini import CFG

    SELECT_POLL = 0.02  # polling time is 20ms

    # WIN32 has no session_fds (multiprocess queues are not polled using
    # select), use a persistently empty set; for WIN32, sessions are always
    # polled for data at every loop.
    WIN32 = sys.platform.lower().startswith('win32')
    session_fds = set()

    log = logging.getLogger('x84.engine')

    if not len(servers):
        raise ValueError("No servers configured for event loop! (ssh, telnet)")

    tap_events = CFG.getboolean('session', 'tap_events')
    check_ban = get_fail2ban_function()
    locks = dict()

    while True:
        # shutdown, close & delete inactive clients,
        for server in servers:
            # bbs sessions that are no longer active on the socket
            # level -- send them a 'kill signal'
            for key, client in server.clients.items()[:]:
                if not client.is_active():
                    kill_session(client, 'socket shutdown')
                    del server.clients[key]
            # on-connect negotiations that have completed or failed.
            # delete their thread instance from further evaluation
            for thread in [_thread for _thread in server.threads
                           if _thread.stopped][:]:
                server.threads.remove(thread)

        check_r = list()
        for server in servers:
            check_r.append(server.server_socket.fileno())
            check_r.extend(server.client_fds())
        if not WIN32:
            # WIN32's IPC is not done using sockets, so it
            # is not possible to use select.select() on them
            session_fds = get_session_output_fds(servers)
            check_r.extend(session_fds)

        # We'd like to use timeout 'None', but the registration of
        # a new client in terminal.start_process surprises us with new
        # file descriptors for the session i/o.  Unless we loop for
        # additional `session_fds', a connecting client would block.
        try:
            ready_r, _, _ = select.select(check_r, [], [], SELECT_POLL)
        except select.error as err:
            # more than likely EBADF (9, 'Bad file descriptor'), it would seem
            # the socket we've just decided to poll has just gone bad.
            log.debug('continue after select.error: {0}'.format(err))
            continue

        for fd in ready_r:
            # see if any new tcp connections were made
            server = find_server(servers, fd)
            if server is not None:
                accept(log, server, check_ban)

        # receive new data from tcp clients.
        client_recv(servers, ready_r, log)
        terms = get_terminals()

        # receive new data from session terminals
        if WIN32 or set(session_fds) & set(ready_r):
            try:
                session_recv(locks, terms, log, tap_events)
            except IOError as err:
                # if the ipc closes while we poll, warn and continue
                log.warn(err)

        # send tcp data to clients
        client_send(terms, log)

        # send session data, poll for user-timeout and disconnect them
        session_send(terms)


if __name__ == '__main__':
    exit(main())



""" System info script for x/84. """


def main():
    """ Main procedure. """
    # pylint: disable=R0914,W0141,R0912
    #         Too many local variables
    #         Used builtin function 'map'
    #         Too many branches
    from x84.bbs import getsession, getterminal, echo, getch, syncterm_setfont
    from x84.engine import __url__ as url
    import platform
    import random
    import sys
    import os
    session, term = getsession(), getterminal()
    session.activity = 'System Info'
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'plant.ans',)
    # pylint: disable=W0633
    #         Attempting to unpack a non-sequence defined at line 1160 of
    #         platform
    system, _, release, _, machine, _ = platform.uname()

    body = [u'authors:',
            u'Johannes Lundberg',
            u'Jeffrey Quast',
            u'Wijnand Modderman-Lenstra',
            u'',
            u'artwork:',
            u'hellbeard!impure',
            u'\r\n',
            u'system: %s %s %s' % (system, release, machine),
            u'software: x/84',
            url,
            u'\r\n',
            (platform.python_implementation() + u' '
                + '-'.join(map(str, sys.version_info[3:])))
            + u' ' + (platform.python_version()
                      if hasattr(platform, 'python_implementation')
                      else u'.'.join(map(str, sys.version_info[:3]))),
            ]
    melt_colors = (
        [term.normal]
        + [term.bold_blue] * 3
        + [term.red] * 4
        + [term.bold_red]
        + [term.bold_white]
        + [term.normal] * 6
        + [term.blue] * 2
        + [term.bold_blue]
        + [term.bold_white]
        + [term.normal])
    art = open(artfile).read().decode('cp437_art') \
        if os.path.exists(artfile) else u''
    otxt = list(art.splitlines())
    for num, line in enumerate(body):
        while num > len(otxt):
            otxt += [u'', ]
        otxt[num] = otxt[num][:int(term.width / 2.5)] + u' ' + line
    width = max([term.length(line) for line in otxt])
    height = len(otxt)
    num_stars = int((term.width * term.height) * .005)
    stars = dict([(n, (random.choice('\\|/-'),
                       float(random.choice(range(term.width))),
                       float(random.choice(range(term.height)))))
                  for n in range(num_stars)])
    melting = {}
    show_star = False
    tm_out, tm_min, tm_max, tm_step = 0.08, 0.01, 2.0, .01
    wind = (0.7, 0.1, 0.01, 0.01)

    def refresh():
        """ Refresh screen and return top-left (x, y) location. """
        # set syncterm font to cp437
        if term.kind.startswith('ansi'):
            echo(syncterm_setfont('cp437'))
        echo(u'\r\n\r\n')
        if term.width < width:
            echo(u''.join((
                term.move(term.height, 0),
                u'\r\n\r\n',
                term.bold_red + 'screen too thin! (%s/%s)' % (
                    term.width, width,),
                u'\r\n\r\n',
                u'press any key...',)))
            getch()
            return (None, None)
        if term.height < height:
            echo(u''.join((
                term.move(term.height, 0),
                u'\r\n\r\n',
                term.bold_red + 'screen too short! (%s/%s)' % (
                    term.height, height),
                u'\r\n\r\n',
                u'press any key...',)))
            getch()
            return (None, None)
        xloc = (term.width / 2) - (width / 2)
        yloc = (term.height / 2) - (height / 2)
        echo(u''.join((
            term.normal,
            (u'\r\n' + term.clear_eol) * term.height,
            u''.join([term.move(yloc + abs_y, xloc) + line
                      for abs_y, line in enumerate(otxt)]),)))
        return xloc, yloc

    txt_x, txt_y = refresh()
    if (txt_x, txt_y) == (None, None):
        return

    def char_at_pos(yloc, xloc, txt_y, txt_x):
        """ Return art (y, x) for location """
        return (u' ' if yloc - txt_y < 0 or yloc - txt_y >= height
                or xloc - txt_x < 0 or xloc - txt_x >= len(otxt[yloc - txt_y])
                else otxt[yloc - txt_y][xloc - txt_x])

    def iter_wind(xslope, yslope, xdir, ydir):
        """ An easterly Wind """
        xslope += xdir
        yslope += ydir
        if xslope <= 0.5:
            xdir = random.choice([0.01, 0.015, 0.02])
        elif xslope >= 1:
            xdir = random.choice([-0.01, -0.015, -0.02])
        if yslope <= -0.1:
            ydir = random.choice([0.01, 0.015, 0.02, 0.02])
        elif yslope >= 0.1:
            ydir = random.choice([-0.01, -0.015, -0.02])
        return xslope, yslope, xdir, ydir

    def iter_star(char, xloc, yloc):
        """ Given char and current position, apply wind and return new
        char and new position. """
        if char == '\\':
            char = '|'
        elif char == '|':
            char = '/'
        elif char == '/':
            char = '-'
        elif char == '-':
            char = '\\'
        xloc += wind[0]
        yloc += wind[1]
        if xloc < 1 or xloc > term.width:
            xloc = (1.0 if xloc > term.width
                    else float(term.width))
            yloc = (float(random.choice
                          (range(term.height))))
        if yloc < 1 or yloc > term.height:
            yloc = (1.0 if yloc > term.height
                    else float(term.height))
            xloc = (float(random.choice
                          (range(term.width))))
        return char, xloc, yloc

    def erase(star_idx):
        """ erase old star before moving .. """
        if show_star:
            _, xloc, yloc = stars[star_idx]
            echo(u''.join((term.move(int(yloc), int(xloc)), term.normal,
                           char_at_pos(int(yloc), int(xloc), txt_y, txt_x),)))

    def melt():
        """ Iterate through all stars and phase through melt sequence. """
        def melted(yloc, xloc):
            """ shift melt, delete if disappeared. """
            melting[(yloc, xloc)] -= 1
            if 0 == melting[(yloc, xloc)]:
                del melting[(yloc, xloc)]
        for (yloc, xloc), phase in melting.items():
            echo(u''.join((term.move(yloc, xloc), melt_colors[phase - 1],
                           char_at_pos(yloc, xloc, txt_y, txt_x),)))
            melted(yloc, xloc)

    def draw_star(star, xloc, yloc):
        """ draw star a (x, y) location """
        char = char_at_pos(int(yloc), int(xloc), txt_y, txt_x)
        if char != ' ':
            melting[(int(yloc), int(xloc))] = len(melt_colors)
        if show_star:
            echo(term.move(int(yloc), int(xloc)) + melt_colors[-1] + star)

    with term.hidden_cursor():
        while txt_x is not None and txt_y is not None:
            if session.poll_event('refresh'):
                num_stars = int(num_stars)
                stars = dict([(n, (random.choice('\\|/-'),
                                   float(random.choice(range(term.width))),
                                   float(random.choice(range(term.height)))))
                              for n in range(num_stars)])
                otxt = list(art.splitlines())
                for num, line in enumerate(body):
                    while num > len(otxt):
                        otxt += [u'', ]
                    otxt[num] = (otxt[num][:int(term.width / 2.5)]
                                 + u' ' + line)
                txt_x, txt_y = refresh()
                continue
            inp = getch(tm_out)
            if inp in (term.KEY_UP, 'k'):
                if tm_out >= tm_min:
                    tm_out -= tm_step
            elif inp in (term.KEY_DOWN, 'j'):
                if tm_out <= tm_max:
                    tm_out += tm_step
            elif inp in (term.KEY_LEFT, 'h'):
                if num_stars > 2:
                    num_stars = int(num_stars * .5)
                    stars = dict([(n, (random.choice('\\|/-'),
                                       float(random.choice(range(term.width))),
                                       float(random.choice(range(term.height)))
                                       )) for n in range(num_stars)])
            elif inp in (term.KEY_RIGHT, 'l'):
                if num_stars < (term.width * term.height) / 4:
                    num_stars = int(num_stars * 1.5)
                    stars = dict([(n, (random.choice('\\|/-'),
                                       float(random.choice(range(term.width))),
                                       float(random.choice(range(term.height)))
                                       )) for n in range(num_stars)])
            elif inp in (u'*',) and not show_star:
                show_star = True
            elif inp in (u'*',) and show_star:
                for star in stars:
                    erase(star)
                show_star = False
            elif inp is not None:
                echo(term.move(term.height, 0))
                break
            melt()
            for star_key, star_val in stars.items():
                erase(star_key)
                # pylint: disable=W0142
                #         Used * or ** magic
                stars[star_key] = iter_star(*star_val)
                draw_star(*stars[star_key])
            # pylint: disable=W0142
            #         Used * or ** magic
            wind = iter_wind(*wind)



""" Terminal handler for x/84 """
import contextlib
import logging
import codecs
import sys
from blessed import Terminal as BlessedTerminal

TERMINALS = dict()


class Terminal(BlessedTerminal):

    """ A thin wrapper over :class:`blessed.Terminal`. """

    _session = None

    def __init__(self, kind, stream, rows, columns):
        """ Class initializer. """
        self._rows = rows
        self._columns = columns
        BlessedTerminal.__init__(self, kind, stream)
        if sys.platform.lower().startswith('win32'):
            self._normal = '\x1b[m'

    @property
    def session(self):
        """ Session associated with this terminal. """
        if self._session is None:
            from x84.bbs.session import getsession
            self._session = getsession()
        return self._session

    def inkey(self, timeout=None, esc_delay=0.35, *_):
        # pylint: disable=C0111
        #         Missing docstring
        try:
            return BlessedTerminal.inkey(self, timeout, esc_delay=0.35)
        except UnicodeDecodeError as err:
            log = logging.getLogger(__name__)
            log.warn('UnicodeDecodeError: {0}'.format(err))
            return u'?'
    inkey.__doc__ = BlessedTerminal.inkey.__doc__

    def set_keyboard_decoder(self, encoding):
        """ Set or change incremental decoder for keyboard input. """
        log = logging.getLogger(__name__)
        try:
            self._keyboard_decoder = codecs.getincrementaldecoder(encoding)()
            self._encoding = encoding
            log.debug('keyboard encoding is {!r}'.format(encoding))
        except Exception as err:
            log.exception(err)

    def kbhit(self, timeout=0, *_):
        # pylint: disable=C0111
        #         Missing docstring
        # pull a value off the input buffer if available,
        val = self.session.read_event('input', timeout)

        # if available, place back into buffer and return True,
        if val is not None:
            self.session.buffer_input(val, pushback=True)
            return True

        # no value available within timeout.
        return False
    kbhit.__doc__ = BlessedTerminal.kbhit.__doc__

    def getch(self):
        # pylint: disable=C0111
        #         Missing docstring
        val = self.session.read_event('input')
        return self._keyboard_decoder.decode(val, final=False)
    getch.__doc__ = BlessedTerminal.getch.__doc__

    def _height_and_width(self):
        # pylint: disable=C0111
        #         Missing docstring
        from blessed.terminal import WINSZ
        return WINSZ(ws_row=self._rows, ws_col=self._columns,
                     ws_xpixel=None, ws_ypixel=None)
    _height_and_width.__doc__ = BlessedTerminal._height_and_width.__doc__

    @contextlib.contextmanager
    def raw(self):
        """ Dummy method yields nothing for blessed compatibility. """
        yield

    @contextlib.contextmanager
    def cbreak(self):
        """ Dummy method yields nothing for blessed compatibility. """
        yield

    @property
    def is_a_tty(self):
        """ Dummy property always returns True. """
        return True


def translate_ttype(ttype):
    """
    Return preferred terminal type given the session-negotiation ttype.

    This provides a kind of coercion; we know some terminals, such as
    SyncTerm report a terminal type of 'ansi' -- however, the author
    publishes a termcap database for 'ansi-bbs' which he instructs
    should be used!  So an ``[system]`` configuration item
    of ``termcap-ansi`` may be set to ``'ansi-bbs'`` to coerce
    such terminals for Syncterm-centric telnet servers -- though I
    would not recommend it.

    Furthermore, if the ttype is (literally) 'unknown', then a
    system-wide default terminal type may be returned, also by
    ``[system]`` configuration option ``termcap-unknown``.
    """
    from x84.bbs import get_ini
    log = logging.getLogger(__name__)

    termcap_unknown = get_ini('system', 'termcap-unknown') or 'ansi'
    termcap_ansi = get_ini('system', 'termcap-ansi') or 'ansi'

    if termcap_unknown != 'no' and ttype == 'unknown':
        log.debug("terminal-type {0!r} => {1!r}"
                  .format(ttype, termcap_unknown))
        return termcap_unknown

    elif (termcap_ansi != 'no' and ttype.lower().startswith('ansi')
          and ttype != termcap_ansi):
        log.debug("terminal-type {0!r} => {1!r}"
                  .format(ttype, termcap_ansi))
        return termcap_ansi

    return ttype


def determine_encoding(env):
    """ Determine and return preferred encoding given session env. """
    from x84.bbs import get_ini
    default_encoding = get_ini(
        section='session', key='default_encoding'
    ) or 'utf8'

    fallback_encoding = {
        'ansi': 'cp437',
        'ansi-bbs': 'cp437',
    }.get(env['TERM'], default_encoding)

    return env.get('encoding', fallback_encoding)


def init_term(writer, env):
    """
    Determine the final TERM and encoding and return a Terminal.

    curses is initialized using the value of 'TERM' of dictionary env,
    as well as a starting window size of 'LINES' and 'COLUMNS'. If the
    terminal-type is of 'ansi' or 'ansi-bbs', then the cp437 encoding
    is assumed; otherwise 'utf8'.

    A blessed-abstracted curses terminal is returned.
    """
    from x84.bbs.ipc import IPCStream
    from x84.bbs import get_ini
    log = logging.getLogger(__name__)
    env['TERM'] = translate_ttype(env.get('TERM', 'unknown'))
    env['encoding'] = determine_encoding(env)
    term = Terminal(kind=env['TERM'],
                    stream=IPCStream(writer=writer),
                    rows=int(env.get('LINES', '24')),
                    columns=int(env.get('COLUMNS', '80')))

    if term.kind is None:
        # the given environment's TERM failed curses initialization
        # because, more than likely, the TERM type was not found.
        termcap_unknown = get_ini('system', 'termcap-unknown') or 'ansi'
        log.debug('terminal-type {0} failed, using {1} instead.'
                  .format(env['TERM'], termcap_unknown))
        term = Terminal(kind=termcap_unknown,
                        stream=IPCStream(writer=writer),
                        rows=int(env.get('LINES', '24')),
                        columns=int(env.get('COLUMNS', '80')))

    log.info("terminal type is {0!r}".format(term.kind))
    return term


class TerminalProcess(object):

    """
    Class record for tracking "terminals".

    Probably of most interest, is that a ``TerminalProcess``
    is an abstract association with a multiprocessing.Process
    sub-process, and its i/o queues (``master_pipes``).

    This is not a really tty, or even a pseudo-tty (pty)!  No
    termios, fnctl, or any terminal driver i/o is performed, it
    is all virtual.

    An instance of this class is stored using :func:`register_tty`
    and removed by :func:`unregister_tty`, and discovered using
    :func:`get_terminals`.
    """

    def __init__(self, client, sid, master_pipes):
        """ Class constructor. """
        from x84.bbs import get_ini
        self.client = client
        self.sid = sid
        (self.master_write, self.master_read) = master_pipes
        self.timeout = get_ini('system', 'timeout') or 0


def flush_queue(queue):
    """
    Flush all data awaiting on the ipc queue.

    Seeks any remaining events in queue, used before closing
    to prevent zombie processes with IPC waiting to be picked up.
    """
    log = logging.getLogger(__name__)
    try:
        while queue.poll():
            event, data = queue.recv()
            if event == 'logger':
                log.handle(data)
    except (EOFError, IOError) as err:
        log.debug(err)


def register_tty(tty):
    """ Register a :class:`TerminalProcess` instance. """
    log = logging.getLogger(__name__)
    log.debug('[{tty.sid}] registered tty'.format(tty=tty))
    TERMINALS[tty.sid] = tty


def unregister_tty(tty):
    """ Unregister a :class:`TerminalProcess` instance. """
    try:
        flush_queue(tty.master_read)
        tty.master_read.close()
        tty.master_write.close()
    except (EOFError, IOError) as err:
        log = logging.getLogger(__name__)
        log.exception(err)
    if tty.client.active:
        # signal tcp socket to close
        tty.client.deactivate()
    del TERMINALS[tty.sid]


def get_terminals():
    """ Returns a list of all terminals as tuples (session-id, ttys). """
    return TERMINALS.items()


def find_tty(client):
    """ Given a client, return a matching tty, or None if not registered. """
    try:
        return next(tty for _, tty in get_terminals() if client == tty.client)
    except StopIteration:
        pass


def kill_session(client, reason='killed'):
    """ Given a client, shutdown its socket and signal subprocess exit. """
    from x84.bbs.exception import Disconnected
    client.shutdown()

    log = logging.getLogger(__name__)
    tty = find_tty(client)
    if tty is not None:
        try:
            tty.master_write.send(('exception', Disconnected(reason),))
        except (EOFError, IOError):
            pass
        log.info('[{tty.sid}] goodbye: {reason}'
                 .format(tty=tty, reason=reason))
        unregister_tty(tty)


def start_process(sid, env, CFG, child_pipes, kind, addrport,
                  matrix_args=None, matrix_kwargs=None):
    """
    A ``multiprocessing.Process`` target.

    :param str sid: string describing session source (IP address & port).
    :param dict env: dictionary of client environment variables
                     (must contain at least ``'TERM'``).
    :param ConfigParser.ConfigParser CFG: bbs configuration
    :param tuple child_pipes: tuple of ``(writer, reader)`` for engine IPC.
    :param str kind: what kind of connection as string, ``'telnet'``,
                     ``'ssh'``, etc.
    :param tuple addrport: ``(client-ip, client-port)`` as string and integer.
    :param tuple matrix_args: optional positional arguments to pass to matrix
                              script.
    :param dict matrix_kwargs: optional keyward arguments to pass to matrix
                               script.
    """
    # pylint: disable=R0913,R0914
    #         Too many arguments (8/5)
    #         Too many local variables (16/15)
    import x84.bbs.ini
    from x84.bbs.ipc import make_root_logger
    from x84.bbs.session import Session

    # CFG must be pickled and sent to child process; on windows systems,
    # fork() does not duplicate that it has been initialized, and requires
    # sending to child process
    x84.bbs.ini.CFG = CFG

    (writer, _) = child_pipes

    # remove any existing log handlers in child process and replace
    # with a new root log handler that sends to x84.bbs.engine over IPC.
    make_root_logger(writer)

    # instantiate and create a new terminal instance given the value
    # of env[TERM], negotiated by protocol. May modify the value of
    # env[TERM] by function translate_ttype
    terminal = init_term(writer=writer, env=env)

    try:
        # instantiate and run session
        kwargs = {
            'terminal': terminal,
            'sid': sid,
            'env': env,
            'child_pipes': child_pipes,
            'kind': kind,
            'addrport': addrport,
            'matrix_args': matrix_args or (),
            'matrix_kwargs': matrix_kwargs or {},
        }
        Session(**kwargs).run()
    finally:
        # signal exit to engine
        try:
            writer.send(('exit', None))
        except IOError as err:
            # ignore [Errno 232] The pipe is being closed,
            # only occurs on win32 platform after early exit
            if err.errno != 232:
                raise


def spawn_client_session(client, matrix_kwargs=None):
    """ Spawn sub-process for connecting client.

    Optional
    """
    from multiprocessing import Process, Pipe
    import x84.bbs.ini

    child_read, master_write = Pipe(duplex=False)
    master_read, child_write = Pipe(duplex=False)
    session_id = '{client.kind}-{client.addrport}'.format(client=client)

    # start sub-process, which will initialize the terminal and
    # begins the 'session' for the connecting client.
    Process(target=start_process, kwargs={
        'sid': session_id,
        'env': client.env,
        'CFG': x84.bbs.ini.CFG,
        'child_pipes': (child_write, child_read),
        'kind': client.kind,
        'addrport': client.addrport,
        'matrix_kwargs': matrix_kwargs,
    }).start()

    # and register its tty and master-side pipes for polling by x84.engine
    register_tty(TerminalProcess(client=client,
                                 sid=session_id,
                                 master_pipes=(master_write, master_read)))


def on_naws(client):
    """
    Callback for telnet NAWS negotiation.

    On a Telnet NAWS sub-negotiation, check if client is yet registered
    in registry, and if so, send a 'refresh' event down the event queue.

    This is ultimately handled by :meth:`x84.bbs.session.Session.buffer_event`.
    """
    for _, tty in get_terminals():
        if client == tty.client:
            columns = int(client.env['COLUMNS'])
            rows = int(client.env['LINES'])
            tty.master_write.send(('refresh', ('resize', (columns, rows),)))
            break
    return True



# -*- coding: utf-8 -*-
""" Session engine for x/84. """

# std imports
import collections
import traceback
import logging
import time
import imp
import sys
import os

# local
from x84.bbs.exception import Disconnected, Goto
from x84.bbs.script_def import Script
from x84.bbs.userbase import User
from x84.bbs.ini import get_ini


#: singleton representing the session connected by current process
SESSION = None


def getsession():
    """ Return :class:`Session` instance of current process. """
    return SESSION


def getterminal():
    """ Return :class:`blessed.Terminal` instance of current session. """
    return getsession().terminal


def goto(script_name, *args, **kwargs):
    """ Change bbs script. Does not return. """
    raise Goto(script_name, *args, **kwargs)


def disconnect(reason=u''):
    """ Disconnect session. Does not return. """
    raise Disconnected(reason,)


def getch(timeout=None):
    """
    A deprecated form of getterminal().inkey().

    This is old behavior -- upstream blessed project does the correct
    thing. please use term.inkey() and see the documentation for
    blessed's inkey() method, it **always** returns unicode, never None,
    and definitely never an integer. However some internal UI libraries
    were built upon getch(), and as such, this remains ...
    """
    # mark deprecate in v2.1; remove entirely in v3.0
    # warnings.warn('getch() is deprecated, use getterminal().inkey()')
    keystroke = getterminal().inkey(timeout)
    if keystroke == u'':
        return None
    if keystroke.is_sequence:
        return keystroke.code
    return keystroke


def gosub(script, *args, **kwargs):
    """ Call bbs script with optional arguments, Returns value. """
    script = Script(name=script, args=args, kwargs=kwargs)
    return getsession().runscript(script)


class Session(object):

    """ A per-process Session. Begins by the :meth:`run`. """

    _encoding = None
    _decoder = None
    _activity = None
    _user = None
    _script_module = None

    def __init__(self, terminal, sid, env, child_pipes, kind, addrport,
                 matrix_args, matrix_kwargs):
        # pylint: disable=R0913
        #        Too many arguments
        """ Instantiate a Session.

        Only one session may be instantiated per process.

        :param blessed.Terminal terminal: interactive terminal associated with
                                          this session.
        :param str sid: session identification string
        :param dict env: transport-negotiated environment variables, should
                         contain at least values for TERM and 'encoding'.
        :param tuple child_pipes: tuple of ``(writer, reader)``.
        :param str kind: transport description string (ssh, telnet)
        :param str addrport: transport ip address and port as string
        :param tuple matrix_args: When non-None, a tuple of positional
                                  arguments passed to the matrix script.
        :param dict matrix_kwargs: When non-None, a dictionary of keyword
                                   arguments passed to the matrix script.
        """
        self.log = logging.getLogger(__name__)

        # pylint: disable=W0603
        #        Using the global statement
        # Manage `SESSION' per-process singleton.
        global SESSION
        assert SESSION is None, 'Only one Session per process allowed'
        SESSION = self

        # public attributes
        self.terminal = terminal
        self.sid = sid
        self.env = env
        self.writer, self.reader = child_pipes
        self.kind = kind
        self.addrport = addrport

        # initialize keyboard encoding
        terminal.set_keyboard_decoder(env['encoding'])

        # private attributes
        self._script_stack = [
            # Initialize the "script stack" with the matrix script
            # using the default configuration value of `script' for all
            # connections, but preferring `script_{kind}', where `kind'
            # may be `telnet', `ssh', or any kind of supporting transport.
            Script(
                name=get_ini('matrix', 'script_{self.kind}'.format(self=self)
                             ) or get_ini('matrix', 'script'),
                args=matrix_args,
                kwargs=matrix_kwargs)
        ]
        self._connect_time = time.time()
        self._last_input_time = time.time()
        self._node = None

        # create event buffer
        self._buffer = dict()

    def to_dict(self):
        """ Dictionary describing this session. """
        retval = {
            attr: getattr(self, attr)
            for attr in (
                'last_input_time',
                'connect_time',
                'encoding',
                'activity',
                'idle',
                'node',
                'pid',
                'sid',
            )
        }
        retval.update(dict(
            handle=self.user.handle,
            term_kind=self.terminal.kind,
            term_width=self.terminal.width,
            term_height=self.terminal.height,
            current_script=self.current_script.name,
        ))
        return retval.copy()

    @property
    def duration(self):
        """ Seconds elapsed since connection began (as float). """
        return time.time() - self.connect_time

    @property
    def connect_time(self):
        """ Time of session start (as float). """
        return self._connect_time

    @property
    def last_input_time(self):
        """ Time of last keypress (as epoch, float). """
        return self._last_input_time

    @property
    def idle(self):
        """ Seconds elapsed since last keypress as float.  """
        return time.time() - self.last_input_time

    @property
    def activity(self):
        """ Current session activity.

        This is arbitrarily set by session scripts.

        This also updates xterm titles, and is globally broadcasted
        as a "current activity" in the Who's online script, for example.
        """
        return self._activity or u'<uninitialized>'

    @activity.setter
    def activity(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        if self._activity != value:
            self.log.debug('activity=%s', value)
            self._activity = value

            if (self.terminal.kind.startswith('xterm') or
                    self.terminal.kind.startswith('rxvt')):
                self.write(u'\x1b]2;{0}\x07'.format(value))

    @property
    def user(self):
        """ :class:`User` instance of this session. """
        return self._user or User()

    @user.setter
    def user(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self.log.info("user {!r} -> {!r}".format(self._user, value.handle))
        self._user = value

    @property
    def encoding(self):
        """ Session encoding, both input and output. """
        return self.env.get('encoding', 'utf8')

    @encoding.setter
    def encoding(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        if value != self.encoding:
            self.log.debug('session encoding {0} -> {1}'
                           .format(self.encoding, value))
            self.env['encoding'] = value
            getterminal().set_keyboard_decoder(value)

    @property
    def pid(self):
        """ Process ID of this session (int). """
        return os.getpid()

    @property
    def tap_input(self):
        """ Whether keyboard input should be logged (bool). """
        return get_ini('session', 'tap_input', getter='getboolean')

    @property
    def tap_output(self):
        """ Whether screen output should be logged (bool). """
        return get_ini('session', 'tap_output', getter='getboolean')

    @property
    def show_traceback(self):
        """ Whether traceback errors should be displayed to user (bool). """
        return get_ini('system', 'show_traceback', getter='getboolean')

    @property
    def script_path(self):
        """ Base filepath folder for all scripts. """
        val = get_ini('system', 'scriptpath')
        # ensure folder exists
        assert os.path.isdir(val), (
            'configuration section [system], value scriptpath: '
            'not a folder: {!r}'.format(val))
        return val

    @property
    def current_script(self):
        """ The current script being executed. """
        self.value = 1
        if len(self._script_stack):
            return self._script_stack[-1]
        return None

    @property
    def script_module(self):
        """ Base python module instance for userland scripts. """
        if self._script_module is None:
            # load default/__init__.py as 'default',
            folder_name = os.path.basename(self.script_path)

            # put it in sys.path for relative imports
            if self.script_path not in sys.path:
                sys.path.insert(0, self.script_path)

            # discover import path to __init__.py, store result
            lookup = imp.find_module('__init__', [self.script_path])
            self._script_module = imp.load_module(folder_name, *lookup)
            self._script_module.__path__ = self.script_path
        return self._script_module

    @property
    def node(self):
        """
        Unique numeric constant for this session.

        This makes it simpler to refer to users who are online, instead
        of by their full session-id (such as telnet-92.32.10.132:57331)
        one can simply refer to node #1, etc..
        """
        if self._node is not None:
            return self._node

        for node in range(1, 65534):
            event = 'lock-%s/%d' % ('node', node)
            self.send_event(event, ('acquire', None))
            data = self.read_event(event)
            if data is True:
                self._node = node
                return self._node

    def __error_recovery(self):
        """ Recover from general exception in script. """
        if 0 != len(self._script_stack):
            # recover from exception
            fault = self._script_stack.pop()
            prefix = u'stop'
            if len(self._script_stack):
                # scripts remaining on the script_stack, resume the script that
                # called us. Make sure your calling script queries for input or
                # some other decision before chaining a gosub(), or you could
                # end up in an infinite loop of gosub() followed by a crash (!)
                resume = self.current_script
                prefix = u'resume {resume.name}'.format(resume=resume)

            # display error to local log handler and to the user,
            msg = (u'{prefix} after general exception in {fault.name}'
                   .format(prefix=prefix, fault=fault))
            self.log.error(msg)
            self.write(u'\r\n\r\n{msg}\r\n'.format(msg=msg))

            # give time for exception to write down the IPC queue before
            # continuing or exiting, esp. exiting, otherwise STOP message
            # is not often fully received to the transport.
            time.sleep(2)

    def run(self):
        """
        Begin main execution of session.

        Scripts manipulate control flow of scripts by raising the
        ``Goto`` exception, or the gosub function.
        """
        self.log.info('Begin session on node %s', self.node)
        try:
            while len(self._script_stack):
                self.log.debug('script_stack is {self._script_stack!r}'
                               .format(self=self))
                try:
                    self.runscript(self._script_stack.pop())
                    continue

                except Goto as goto_script:
                    self.log.debug('goto {0}'.format(goto_script.value))
                    self._script_stack = [goto_script.value]
                    continue

                except Disconnected as err:
                    self.log.info('Disconnected: %s', err)
                    return None

                except Exception as err:
                    # Pokemon exception, log and Cc: client, then resume.
                    e_type, e_value, e_tb = sys.exc_info()
                    if self.show_traceback:
                        self.write(self.terminal.normal + u'\r\n')

                    terrs = list()
                    for line in traceback.format_tb(e_tb):
                        for subln in line.split('\n'):
                            terrs.append(subln)
                    terrs.extend(
                        traceback.format_exception_only(e_type, e_value))

                    for e_txt in map(str.rstrip, terrs):
                        self.log.error(e_txt)
                        if self.show_traceback:
                            self.write(e_txt + u'\r\n')

                # recover from general exception
                self.__error_recovery()
            self.log.debug('End of script stack.')
        finally:
            self.close()

    def write(self, ucs, encoding=None):
        """ Write unicode data ``ucs`` to terminal. """
        # do not write empty strings
        if not ucs:
            return
        self.terminal.stream.write(ucs, encoding or self.encoding)

        if self.log.isEnabledFor(logging.DEBUG) and self.tap_output:
            self.log.debug('--> {!r}'.format(ucs))

    def flush_event(self, event):
        """
        Flush and return all data buffered for ``event``.

        :param str event: event name.
        :rtype: list
        """
        flushed = list()
        while True:
            data = self.read_event(event, -1)
            if data is None:
                if 0 != len(flushed):
                    self.log.debug('flushed from %s: %r', event, flushed)
                return flushed
            flushed.append(data)
        return flushed

    def buffer_event(self, event, data=None):
        """
        Buffer and handle IPC data keyed by ``event``.

        :param str event: event name.
        :param data: event data.
        :rtype: bool
        :returns: True if the event was internally handled, and the caller
                  should take no further action.

        Methods internally handled by this method:

        - ``global``: events where the first index of ``data`` is ``AYT``.
          This is sent by other sessions using the ``broadcast`` event, to
          discover "who is online".

        - ``info-req``: Where the first data value is the remote session-id
          that requested it, expecting a return value event of ``info-ack``
          whose data values is a dictionary describing a session.  This is
          an extension of the "who is online" event described above.

        - ``gosub``: Allows one session to send another to a different script,
          this is used by the default board ``chat.py`` for a chat request.
        """
        # exceptions aren't buffered; they are thrown!
        if event == 'exception':
            # pylint: disable=E0702
            #        Raising NoneType while only classes, (..) allowed
            raise data

        # these callback-responsive session events should be handled by
        # another method, or by a configurable 'event: callback' registration
        # system.
        # respond to global 'AYT' requests
        if event == 'global' and data[0] == 'AYT':
            reply_to = data[1]
            self.send_event('route', (
                reply_to, 'ACK',
                self.sid, self.user.handle,))
            return True

        # accept 'gosub' as a literal command to run a new script directly
        # from this buffer_event method.  I'm sure it's fine ...
        if event == 'gosub':
            save_activity = self.activity
            self.log.info('event-driven gosub: {0}'.format(data))
            try:
                self.runscript(Script(*data))
            finally:
                self.activity = save_activity
                # RECURSIVE: we call buffer_event to push-in a duplicate
                # "resize" event, so the script that was interrupted has
                # an opportunity to adjust to the new terminal dimensions
                # (if any), or in the case of subpar clients such as netrunner
                # and syncterm that do not handle "alt screen" correctly,
                # it forces a refresh in the interrupted script, though there
                # aren't any guarantees.
                # Otherwise, it is fine to not require the calling function to
                # refresh -- so long as the target script makes sure(!) to
                # use the "with term.fullscreen()" context manager.
                data = ('resize', (self.terminal.height, self.terminal.width,))
                self.buffer_event('refresh', data)
            return True

        # respond to 'info-req' events by returning pickled session info
        if event == 'info-req':
            self.send_event('route', (
                # data[0] is the session-id to return the reply to.
                data[0], 'info-ack', self.sid, self.to_dict()))
            return True

        if event not in self._buffer:
            # " Once a bounded length deque is full, when new items are added,
            # a corresponding number of items are discarded from the opposite
            # end." -- global events are meant to be disregarded, so a much
            # shorter queue length is used. only the foremost refresh event is
            # important in the case of screen resize.
            self._buffer[event] = collections.deque(
                maxlen={'global': 128,
                        'refresh': 1,
                        }.get(event, 65534))

        # buffer input
        if event == 'input':
            self.buffer_input(data)
            return False

        # buffer only 1 most recent 'refresh' event
        if event == 'refresh':
            if data[0] == 'resize':
                # inherit terminal dimensions values
                (self.terminal._columns, self.terminal._rows) = data[1]

        # buffer all else
        self._buffer[event].appendleft(data)

        return False

    def buffer_input(self, data, pushback=False):
        """
        Receive keyboard input ,``data``, into ``input`` buffer.

        Updates idle time, buffering raw bytes received from telnet client via
        event queue.  Sometimes a script may poll for, and receive keyboard
        data, but wants to push it back in to the top of the stack to be
        decoded by a later call to term.inkey(); in such case, ``pushback``
        should be set.

        :param bytes data: keyboard input data.
        :param bool pushback: whether it should be pushed to front of stack.
        """
        self._last_input_time = time.time()

        if self.log.isEnabledFor(logging.DEBUG) and self.tap_input:
            self.log.debug('<-- {!r}'.format(data))

        if 'input' not in self._buffer:
            # a rare scenario: inkey() causes a first-event of 'input' to
            # be received without buffering from read_events(); then wants
            # to push it back.  Here, too, we must check and construct the
            # input buffer.  It wouldn't be bad to do this on __init__,
            # either.
            self._buffer['input'] = collections.deque(maxlen=65534)
        if pushback:
            self._buffer['input'].appendleft(data)
        else:
            self._buffer['input'].append(data)

    def send_event(self, event, data):
        """
        Send data to IPC output queue in form of (event, data).

        Supported ``event`` strings:

        - ``disconnect``: Session wishes to disconnect.

        - ``logger``: Data is logging record, used by IPCLogHandler.

        - ``output``: Unicode data to write to client.

        - ``global``: Broadcast event to other sessions.

        - ``route``: Send an event to another session.

        - ``db-<schema>``: Request sqlite dict method result.

        - ``db=<schema>``: Request sqlite dict method result as iterable.

        - ``lock-<name>``: Fine-grained global bbs locking.

        :param str event: event name.
        :param data: event data.
        """
        self.writer.send((event, data))

    def poll_event(self, event):
        """
        Non-blocking poll for session event.

        :param str event: an IPC event queue by name, such as ``input``.
        :returns: first matching IPC event data, or ``None``.
        """
        return self.read_event(event, timeout=-1)

    def read_event(self, event, timeout=None):
        """
        Return data for given ``event`` by timeout.

        :param str event: an IPC event queue by name, such as ``input``.
        :param int timeout: Value of ``None`` is blocking (default), ``-1`` is
                            non-blocking poll. All other values are blocking up
                            to value of timeout.
        :returns: first matching IPC event data.  If timeout is specified and
                  no matching IPC event is discovered, ``None`` is returned.
        """
        return self.read_events(events=(event,), timeout=timeout)[1]

    def read_events(self, events, timeout=None):
        """
        Return the first matched IPC data for any event specified by timeout.

        :param tuple events: events to search for, for example
                             ``('input', 'refresh')``.
        :param int timeout: Value of ``None`` is blocking (default), ``-1`` is
                            non-blocking poll. All other values are blocking up
                            to value of timeout.
        :rtype: tuple
        :returns: first matching IPC ``event, data`` tuple, where ``event``
                  matches one of the given ``events``.  If timeout is specified
                  and no matching IPC event is discovered, ``(None, None)`` is
                  returned.
        """
        event, data = self._pop_event_buffer(events)
        if event:
            return (event, data)

        timeleft = lambda cmp_time: (
            None if timeout is None else
            timeout if timeout < 0 else
            timeout - (time.time() - cmp_time))

        # begin scanning for matching `events' up to timeout.
        stime = time.time()
        waitfor = timeleft(stime)
        while waitfor is None or waitfor > 0:
            # ask engine process for new event data,
            poll = min(0.5, waitfor) or 0.01
            if self.reader.poll(poll):
                event, data = self.reader.recv()
                # it is necessary to always buffer an event, as some
                # side-effects may occur by doing so.  When buffer_event
                # returns True, those side-effects caused no data to be
                # buffered, and one should not try to return any data for it.
                if not self.buffer_event(event, data):
                    if event in events:
                        return event, self._buffer[event].pop()
            else:
                event, data = self._pop_event_buffer(events)
                if event is not None:
                    return (event, data)
                elif timeout == -1:
                    return (None, None)
            waitfor = timeleft(stime)
        return (None, None)

    def _pop_event_buffer(self, events):
        """
        Return immediately any event-data already buffered.

        :rtype: tuple
        """
        return next(
            ((_event, self._buffer[_event].pop())
             for _event in events
             if len(self._buffer.get(_event, []))),
            (None, None))

    def runscript(self, script):
        """
        Execute the main() callable of script identified by ``script``.

        :param Script script: target script to execute.
        :returns: the return value of the given script's ``main()`` function.
        """
        self.log.info("runscript {0!r}".format(script.name))
        self._script_stack.append(script)

        # if given a script name such as 'extras.target', adjust the lookup
        # path to be extended by {default_scriptdir}/extras, and adjust
        # script_name to be just 'target'.
        script_relpath = self.script_module.__path__
        lookup_paths = [script_relpath]
        if '.' not in script.name:
            script_name = script.name
        else:
            # build another system path, relative to `script_module'
            remaining, script_name = script.name.rsplit('.', 1)
            _lookup_path = os.path.join(script_relpath, *remaining.split('.'))
            lookup_paths.append(_lookup_path)

        lookup = imp.find_module(script_name, lookup_paths)
        module = imp.load_module(script_name, *lookup)

        # ensure main() function exists!
        if not hasattr(module, 'main'):
            raise RuntimeError("script {0}, module {1}: main() not found."
                               .format(script_name, module))
        if not callable(module.main):
            raise RuntimeError("script {0}, module {1}: {2} not callable."
                               .format(script_name, module, module.main))

        # capture the return value of the script and return
        # to the caller -- so value = gosub('my_game') can retrieve
        # the return value of its main() function.
        value = module.main(*script.args, **script.kwargs)

        # remove the current script from the script stack, since it has
        # finished executing.
        self._script_stack.pop()

        return value

    def close(self):
        """ Close session, currently releases ``node`` lock.. """
        if self._node is not None:
            self.send_event(
                event='lock-node/%d' % (self._node),
                data=('release', None))



""" Userlister for x/84. """
# std
import collections
import time
import os

# local
from x84.bbs import getsession, getterminal
from x84.bbs import echo, timeago, get_ini
from x84.bbs import get_user, list_users

from common import display_banner, prompt_pager

#: filepath to folder containing this script
here = os.path.dirname(__file__)

#: filepath to artfile displayed for this script
art_file = get_ini(
    section='userlist', key='art_file'
) or os.path.join(here, 'art', 'userlist.ans')

#: encoding of artfile
art_encoding = get_ini(
    section='userlist', key='art_encoding'
) or 'cp437'

#: preferred fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='userlist', key='syncterm_font'
) or 'topaz'

#: maximum length of user handles
username_max_length = get_ini(
    section='nua', key='max_user', getter='getint'
) or 10

#: maximum length of location, hard-coded to match art_file.
#: maximum length of user 'location' field
location_max_length = get_ini(
    section='nua', key='max_location', getter='getint'
) or 25


def iter_userlist():
    handles = sorted(list_users(), key=unicode.lower)
    timenow = time.time()
    user_record = collections.namedtuple('userlist', [
        'handle', 'location', 'timeago'])
    return (user_record(handle=user.handle,
                        location=user.location,
                        timeago=timenow - user.lastcall)
            for user in (get_user(handle) for handle in handles))


def main():
    session, term = getsession(), getterminal()
    session.activity = 'Viewing Userlist'

    colors = {'highlight': term.red,
              'lowlight': term.green, }

    line_no = display_banner(filepattern=art_file, encoding=art_encoding)

    def make_header(fmt):
        return fmt.format(
            handle=term.bold('handle'.ljust(username_max_length)),
            location=term.bold('location'.ljust(location_max_length)),
            lastcall=term.bold('time ago').ljust(8))

    userlist_fmt = u'| {handle} | {location} | {lastcall} |'

    header = make_header(userlist_fmt)
    header_length = term.length(header)

    # for smaller screens, remove 'location' field.
    if header_length > term.width:
        userlist_fmt = u'| {handle} | {lastcall} |'
        header = make_header(userlist_fmt)
        header_length = term.length(header)

    userlist = [header] + ['-' * header_length]

    for _ur in iter_userlist():
        location_txt = u''
        if 'location' in userlist_fmt:
            location_txt = colors['lowlight'](
                _ur.location.ljust(location_max_length))
        timeago_txt = timeago(_ur.timeago).ljust(8)
        handle_txt = _ur.handle.ljust(username_max_length)

        userlist.append(userlist_fmt.format(
            handle=handle_txt,
            location=location_txt,
            lastcall=timeago_txt))

    echo(u'\r\n')

    # display users, using a command-prompt pager.
    prompt_pager(content=userlist,
                 line_no=line_no + 1,
                 colors={'highlight': term.red,
                         'lowlight': term.green,
                         },
                 width=80, breaker=None)



""" SSH server for x84. """

from __future__ import absolute_import

# standard
import threading
import logging
import socket
import array
import errno
import time
import os

# local
from x84.bbs.exception import Disconnected
from x84.bbs.userbase import (
    check_new_user,
    check_bye_user,
    check_anonymous_user,
    check_user_password,
    check_user_pubkey,
)

from x84.terminal import spawn_client_session, on_naws
from x84.client import BaseClient, BaseConnect
from x84.server import BaseServer
from x84.sftp import X84SFTPServer

# 3rd-party
import paramiko
import paramiko.py3compat


class SshClient(BaseClient):

    """A remote Ssh Client, instantiated from SshServer. """

    def __init__(self, sock, address_pair, on_naws=None):
        super(SshClient, self).__init__(sock, address_pair, on_naws)

        # Becomes the ssh transport
        self.transport = None

        # Becomes the ssh session channel
        self.channel = None

        # the session kind may be transposed into 'sftp'
        # if such subsystem is enabled and connected.
        self.kind = 'ssh'

    def shutdown(self):
        """
        Shutdown and close socket.

        Called by event loop after client is marked by :meth:`deactivate`.
        """
        self.active = False
        if self.channel is not None:
            try:
                # """ only close the pipe when the user explicitly closes the
                # channel. otherwise they will get unpleasant surprises. (and
                # do it before checking self.closed, since the remote host may
                # have already closed the connection.)
                self.channel.close()
            except Exception as err:
                self.log.debug('{self.addrport}: channel close '
                               '{self.__class__.__name__}: {err}'
                               .format(self=self, err=err))
            try:
                # If ``how`` is 2, further sends and receives are disallowed.
                # This closes the stream in one or both directions.
                self.channel.shutdown(how=2)
            except EOFError:
                pass
            finally:
                self.log.debug('{self.addrport}: channel shutdown '
                               '{self.__class__.__name__}'.format(self=self))

        if self.transport is not None and self.transport.is_active():
            self.transport.close()
            self.log.debug('{self.addrport}: transport shutdown '
                           '{self.__class__.__name__}'.format(self=self))

    def is_active(self):
        """ Whether this connection is active (bool). """
        if self.transport is None or self.channel is None:
            # still connecting/negotiating, return our static
            # value (which is True, unless shutdown was called)
            return self.active
        return self.transport.is_active()

    def send_ready(self):
        """ Whether any data is buffered for delivery. """
        if self.channel is None:
            # channel has not yet been negotiated
            return False
        return self.send_buffer.__len__() and self.channel.send_ready()

    def _send(self, send_bytes):
        """
        Sends ``send_bytes`` to ssh channel, returning number of bytes sent.

        Caller must re-buffer bytes not sent.
        :raises Disconnected: on socket send error (client disconnect).
        """
        try:
            return self.channel.send(send_bytes)
        except EOFError:
            raise Disconnected('EOFError')
        except socket.error as err:
            if err[0] == errno.EDEADLK:
                self.log.debug('{self.addrport}: {err} (bandwidth exceed)'
                               .format(self=self, err=err))
                return 0
            raise Disconnected('socket error: {err}'.format(err=err))

    def send(self):
        """
        Send any data buffered and return number of bytes send.

        :raises Disconnected: client has disconnected (cannot write to socket).
        """
        if not self.send_ready():
            self.log.warn('send() called on empty buffer')
            return 0

        ready_bytes = bytes(''.join(self.send_buffer))
        self.send_buffer = array.array('c')

        sent = self._send(ready_bytes)
        if sent < len(ready_bytes):
            # re-buffer data that could not be pushed to socket;
            self.send_buffer.fromstring(ready_bytes[sent:])
        return sent

    def recv_ready(self):
        """ Whether data is awaiting on the ssh channel.  """
        if self.channel is None or self.kind == 'sftp':
            # very strange for SFTP, all i/o is handled by paramiko's event
            # loop and the various callback handlers of x84/sftp.py.  We always
            # return False.  If we enable this, we'll find "input" received
            # into matrix_sftp, which is raw protocol bytes that we should not
            # concern ourselves with.
            return False
        return self.channel.recv_ready()

    def socket_recv(self):
        """
        Receive data from socket, returns number of bytes received.

        :raises Disconnect: client has disconnected.
        :rtype: int
        """
        recv = 0
        try:
            data = self.channel.recv(self.BLOCKSIZE_RECV)
            recv = len(data)
            if 0 == recv:
                raise Disconnected('Closed by client (EOF)')
        except socket.error as err:
            raise Disconnected('socket error: {err}'.format(err=err))
        self.bytes_received += recv
        self.last_input_time = time.time()
        self.recv_buffer.fromstring(data)
        return recv


class ConnectSsh(BaseConnect):

    """
    SSH protocol on-connect handler (in thread).

    Takes care of the (initial) handshake, authentication,
    terminal, and session setup, ultimately spawning a
    process by :func:`spawn_client_session` on success.
    """

    #: time to periodically poll for negotiation completion.
    TIME_POLL = 0.05

    #: time to give up awaiting session negotiation.
    TIME_WAIT_STAGE = 30

    def __init__(self, client, server_host_key, on_naws=None):
        """
        Class constructor.

        :param ssh.SshClient client: an SshClient instance.
        :param paramiko.RSAKey server_host_key: an RSAKey instance.
        """
        self.server_host_key = server_host_key
        self.on_naws = on_naws
        super(ConnectSsh, self).__init__(client)

    def run(self):
        """ Accept new Ssh connect in thread. """
        try:
            self.client.transport = paramiko.Transport(self.client.sock)
            self.client.transport.load_server_moduli()
            self.client.transport.add_server_key(self.server_host_key)
            ssh_session = SshSessionServer(client=self.client)
            from x84.bbs import get_ini
            if get_ini(section='sftp', key='enabled', getter='getboolean'):
                self.client.transport.set_subsystem_handler(
                    'sftp', paramiko.SFTPServer, X84SFTPServer,
                    ssh_session=ssh_session)

            def detected():
                """ Whether shell or SFTP session has been detected. """
                return (ssh_session.shell_requested.isSet() or
                        ssh_session.sftp_requested.isSet())

            self.client.transport.start_server(server=ssh_session)

            st_time = time.time()
            while self._timeleft(st_time):
                self.client.channel = self.client.transport.accept(1)
                if self.client.channel is not None:
                    break
                if not self.client.transport.is_active():
                    self.log.debug('{client.addrport}: transport closed.'
                                   .format(client=self.client))
                    self.client.deactivate()
                    return
            else:
                self.log.debug('{client.addrport}: no channel requested'
                               .format(client=self.client))
                self.client.deactivate()
                return

            first_log = False
            while not detected() and self._timeleft(st_time):
                if not first_log:
                    self.log.debug('{client.addrport}: waiting for '
                                   'shell or subsystem request.'
                                   .format(client=self.client))
                    first_log = True

                if not self.client.is_active():
                    self.log.debug('{client.addrport}: transport closed '
                                   'while waiting for shell or subsystem '
                                   'request.'
                                   .format(client=self.client))
                    self.client.deactivate()
                    return
                time.sleep(self.TIME_POLL)

            if detected():
                matrix_kwargs = {attr: getattr(ssh_session, attr)
                                 for attr in ('anonymous', 'new', 'username',)}
                return spawn_client_session(client=self.client,
                                            matrix_kwargs=matrix_kwargs)

        except (paramiko.SSHException, socket.error) as err:
            self.log.debug('{client.addrport}: connection closed: {err}'
                           .format(client=self.client, err=err))
        except EOFError:
            self.log.debug('{client.addrport}: EOF from client'
                           .format(client=self.client))
        except Exception as err:
            self.log.debug('{client.addrport}: connection closed: {err}'
                           .format(client=self.client, err=err))
        else:
            self.log.debug('{client.addrport}: shell not requested'
                           .format(client=self.client))
        finally:
            self.stopped = True
        self.client.deactivate()

    def _timeleft(self, st_time):
        """
        Whether time elapsed since ``st_time`` is below ``TIME_WAIT_STAGE``.

        Returns True when difference of current time and st_time is below
        TIME_WAIT_STAGE, and the ``stopped`` class attribute has not yet
        been set (such as during server shutdown).
        """
        return bool(not self.stopped and
                    time.time() - st_time < self.TIME_WAIT_STAGE)


class SshSessionServer(paramiko.ServerInterface):

    """
    SSH on-connect Session Server interface.

    Methods of this class are callbacks from Paramiko's primary thread,
    generally returning whether to accept or deny authentication and
    sub-system requests.
    """

    def __init__(self, client):
        self.shell_requested = threading.Event()
        self.sftp_requested = threading.Event()
        self.log = logging.getLogger(__name__)
        self.client = client

        # to be checked by caller
        self.new = False
        self.anonymous = False
        self.username = None
        self.sftp = False

    def check_channel_request(self, kind, chanid):
        # pylint: disable=W0613
        #         Unused argument 'chanid'
        if kind == 'session':
            self.log.debug('channel request granted, kind={0}'.format(kind))
            return paramiko.OPEN_SUCCEEDED
        self.log.debug('channel request denied, kind={0}'.format(kind))
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_auth_password(self, username, password):
        """ Return success/fail for username and password. """
        self.username = username.strip()

        if self.check_account_noverify(username):
            self.log.info('any password accepted for system-enabled '
                          'account, {0!r}'.format(username))
            return paramiko.AUTH_SUCCESSFUL
        if check_user_password(username, password):
            self.log.info('password accepted for user {0!r}.'.format(username))
            return paramiko.AUTH_SUCCESSFUL

        self.log.info('password rejected for user {0!r}.'.format(username))
        return paramiko.AUTH_FAILED

    def check_auth_publickey(self, username, public_key):
        self.username = username.strip()
        if self.check_account_noverify(username):
            self.log.info('any pubkey accepted for system-enabled '
                          'account, {0!r}'.format(username))
            return paramiko.AUTH_SUCCESSFUL
        elif check_user_pubkey(username, public_key):
            self.log.info('pubkey accepted for user {0!r}.'.format(username))
            return paramiko.AUTH_SUCCESSFUL
        self.log.debug('pubkey denied for user {0!r}.'
                       .format(username))
        return paramiko.AUTH_FAILED

    def check_account_noverify(self, username):
        """ Return success/fail for system-enabled accounts.

        For some usernames, such as 'new' or 'anonymous', a correct
        password or public key is not required -- any will do. We return
        True if ``username`` is one of these configurable account names
        and if it is enabled.

        This method has two side effects, it may set the instance
        attribute ``new_user`` or ``anonymous`` to True if it is enabled
        by configuration and the username is of their matching handles.
        """
        if check_new_user(username):
            # if allowed, allow new@, etc. to apply for an account.
            self.new = True
            self.log.debug('accepted without authentication, {0!r}: '
                           'it is an alias for new user application.'
                           .format(username))
            return True

        elif check_bye_user(username):
            # not allowed to login using bye@, logoff@, etc.
            self.log.debug('denied user, {0!r}: it is an alias for logoff'
                           .format(username))
            return False

        elif check_anonymous_user(username):
            # if enabled, allow ssh anonymous@, root@, etc.
            self.log.debug('anonymous user, {0!r} accepted by configuration.'
                           .format(username))
            self.anonymous = True
            return True

        return False

    def get_allowed_auths(self, username):
        # pylint: disable=W0613
        #         Unused argument 'username'
        return 'password,publickey'

    def check_channel_shell_request(self, channel):
        # pylint: disable=W0613
        #         Unused argument 'channel'
        self.log.debug('ssh channel granted.')
        self.shell_requested.set()
        return True

    def check_channel_subsystem_request(self, channel, name):
        from x84.bbs import get_ini
        if name == 'sftp':
            if get_ini(section='sftp', key='enabled', getter='getboolean'):
                self.client.kind = 'sftp'
                self.sftp_requested.set()
                self.sftp = True

        return paramiko.ServerInterface.check_channel_subsystem_request(
            self, channel, name)

    def check_channel_pty_request(self, channel, term, width, height, *_):
        # pylint: disable=W0613
        #         Unused argument 'channel'
        self.client.env['TERM'] = term
        self.client.env['LINES'] = str(height)
        self.client.env['COLUMNS'] = str(width)
        return True

    def check_channel_window_change_request(self, channel, width, height, *_):
        # pylint: disable=W0613
        #         Unused argument 'channel'
        self.client.env['LINES'] = str(height)
        self.client.env['COLUMNS'] = str(width)
        if self.client.on_naws is not None:
            self.client.on_naws(self.client)
        return True

    def check_channel_env_request(self, channel, name, value):
        # pylint: disable=W0613
        #         Unused argument 'channel'
        self.log.debug('env request: [{0}] = {1}'.format(name, value))
        self.client.env[name] = value
        return True


class SshServer(BaseServer):

    """ SSH Server, tracking connecting clients for send/recv management. """

    client_factory = SshClient
    client_factory_kwargs = dict(on_naws=on_naws)
    connect_factory = ConnectSsh

    @classmethod
    def connect_factory_kwargs(cls, instance):
        return dict(server_host_key=instance.host_key)

    # Dictionary of active clients, (file descriptor, SshClient,)
    clients = {}

    def __init__(self, config):
        """ Class initializer. """
        self.log = logging.getLogger(__name__)
        self.config = config
        self.address = config.get('ssh', 'addr')
        self.port = config.getint('ssh', 'port')

        if self.config.has_option('ssh', 'HostKey'):
            filename = config.get('ssh', 'HostKey')
        else:
            filename = os.path.join(
                os.path.expanduser((config.get('system', 'datapath'))),
                'ssh_host_rsa_key')

        if not os.path.exists(filename):
            self.host_key = self.generate_host_key(filename)
        else:
            self.host_key = paramiko.RSAKey(filename=filename)
            self.log.debug('Loaded host key {0}'.format(filename))

        # bind
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        try:
            self.server_socket.bind((self.address, self.port))
            self.server_socket.listen(self.LISTEN_BACKLOG)
        except socket.error as err:
            self.log.error('Unable to bind {self.address}:{self.port}, {err}'
                           .format(self=self, err=err))
            exit(1)
        self.log.info('ssh listening on {self.address}:{self.port}/tcp'
                      .format(self=self))

    def generate_host_key(self, filename):
        """ Generate server host key to local filepath ``filename``. """
        from paramiko import RSAKey

        bits = 4096
        if self.config.has_option('ssh', 'HostKeyBits'):
            bits = self.config.getint('ssh', 'HostKeyBits')

        # generate private key and save,
        self.log.info('Generating {bits}-bit RSA public/private keypair.'
                      .format(bits=bits))
        priv_key = RSAKey.generate(bits=bits)
        priv_key.write_private_key_file(filename, password=None)
        self.log.debug('{filename} saved.'.format(filename=filename))

        # save public key,
        pub = RSAKey(filename=filename, password=None)
        with open('{0}.pub'.format(filename,), 'w') as fp:
            fp.write("{0} {1}".format(pub.get_name(), pub.get_base64()))
        self.log.debug('{filename}.pub saved.'.format(filename=filename))
        return priv_key

    def client_fds(self):
        """ Return list of client file descriptors to poll for read/write. """
        return [_client.channel.fileno() for _client in self.clients.values()
                if _client.channel is not None]

    def clients_ready(self, ready_fds=None):
        """
        Return a list of clients with data ready to be receive.

        The ``ready_fds`` parameter is ignored by the SSH Server.
        """
        return [client for client in self.clients.values()
                if client.recv_ready()]



#!/usr/bin/env python
def main():
    """
    Displays all known key capabilities that may match the terminal.
    As each key is pressed on input, it is lit up and points are scored.
    """
    try:
        from x84.bbs import getterminal, echo
        term = getterminal()
    except (ImportError, AttributeError):
        from blessed import Terminal
        import sys
        term = Terminal()

        def echo(text):
            sys.stdout.write(u'{}'.format(text))
            sys.stdout.flush()

    echo(u''.join((term.normal,
                   term.height * u'\r\n',
                   term.home,
                   term.clear_eos)))

    with term.raw():
        inp = u''
        echo(u'Press Q to exit.\r\n')
        while inp.upper() != 'Q':
            inp = term.inkey(timeout=10.0)
            disp_inp = inp.__str__() if inp.is_sequence else inp
            echo(u'{0!r}: code={1!r} name={2!r}\r\n'
                 .format(disp_inp, inp.code, inp.name))

if __name__ == '__main__':
    main()



"""
fail2ban module for x/84.

To enable, add to default.ini::

    [fail2ban]
    enabled = yes

The following options are available, but not required:

- ``ip_blacklist``: space-separated list of IPs on permanent blacklist.
- ``ip_whitelist``: space-separated list of IPs to always allow.
- ``max_attempted_logins``: max no. of logins allowed for given time window
- ``max_attempted_logins_window``: the length (in seconds) of the time window
  for which logins will be tracked (sliding scale).
- ``initial_ban_length``: ban length (in seconds) when an IP is blacklisted.
- ``ban_increment_length``: amount of time (in seconds) to add to a ban on
  subsequent login attempts
"""

# std imports
import logging
import time

# globals
BANNED_IP_LIST, ATTEMPTED_LOGINS = dict(), dict()


def get_fail2ban_function():
    """
    Return a function used to ban aggressively-connecting clients.

    This is analogous to the 'fail2ban' utility, for example, telnet
    or ssh connect scanners.

    Returns a function which may be passed an IP address, returning True
    if the connection from address ``ip`` should be accepted.

    :return: function accepting ip address, returning boolean
    :rtype: callable
    """
    # local imports
    from x84.bbs import get_ini

    if not get_ini(section='fail2ban', key='enabled', getter='getboolean'):
        return lambda ip: True

    # configuration
    ip_blacklist = get_ini(section='fail2ban',
                           key='ip_blacklist',
                           split=True)

    ip_whitelist = get_ini(section='fail2ban',
                           key='ip_whitelist',
                           split=True)

    max_attempted_logins = get_ini(
        section='fail2ban',
        key='max_attempted_logins',
        getter='getint'
    ) or 3

    max_attempted_logins_window = get_ini(
        section='fail2ban',
        key='max_attempted_logins_window',
        getter='getint'
    ) or 30

    initial_ban_length = get_ini(
        section='fail2ban',
        key='initial_ban_length',
        getter='getint'
    ) or 360

    ban_increment_length = get_ini(
        section='fail2ban',
        key='ban_increment_length',
        getter='getint'
    ) or 360

    def wrapper(ip):
        """ Inner wrapper function. """
        log = logging.getLogger(__name__)

        # pylint: disable=W0602
        #         Using global for 'BANNED_IP_LIST' but no assignment is done
        global BANNED_IP_LIST, ATTEMPTED_LOGINS

        now = int(time.time())

        # check to see if IP is blacklisted
        if ip in ip_blacklist:
            log.debug('Blacklisted IP rejected: {ip}'.format(ip=ip))
            return False

        # check to see if IP is banned
        elif ip in BANNED_IP_LIST:
            # expired?
            if now > BANNED_IP_LIST[ip]:
                # expired ban; remove it
                del BANNED_IP_LIST[ip]
                ATTEMPTED_LOGINS[ip] = {
                    'attempts': 1,
                    'expiry': now + max_attempted_logins_window
                }
                log.debug('Banned IP expired: {ip}'.format(ip=ip))
            else:
                # increase the expiry and kick them out
                BANNED_IP_LIST[ip] += ban_increment_length
                log.debug('Banned IP rejected: {ip}'.format(ip=ip))
                return False

        # check num of attempts, ban if exceeded max
        elif ip in ATTEMPTED_LOGINS:
            if now > ATTEMPTED_LOGINS[ip]['expiry']:
                # window closed; start over
                record = ATTEMPTED_LOGINS[ip]
                record['attempts'] = 1
                record['expiry'] = now + max_attempted_logins_window
                ATTEMPTED_LOGINS[ip] = record
                log.debug('Attempt outside of expiry window')
            elif ATTEMPTED_LOGINS[ip]['attempts'] > max_attempted_logins:
                # max # of attempts reached
                del ATTEMPTED_LOGINS[ip]
                BANNED_IP_LIST[ip] = now + initial_ban_length
                log.warn('Exceeded maximum attempts; banning {ip}'
                         .format(ip=ip))
                return False
            else:
                # extend window
                record = ATTEMPTED_LOGINS[ip]
                record['attempts'] += 1
                record['expiry'] += max_attempted_logins_window
                ATTEMPTED_LOGINS[ip] = record
                log.debug('Window extended')

        # log attempted login
        elif ip not in ip_whitelist:
            log.debug('First attempted login for this window')
            ATTEMPTED_LOGINS[ip] = {
                'attempts': 1,
                'expiry': now + max_attempted_logins_window,
            }
        return True

    return wrapper



""" Editor package for x/84. """
# TODO(jquast): upstream, the blessed project has a ticket to determine
# the current cursor position, https://github.com/jquast/blessed/issues/19
# once implemented, LineEditor can be used with absolute positioning.
# std imports
#
# TODO(jquast): Surely somebody has written a readline-like implementation
# for python that we could use instead.  We cannot use readline directly
# due to its C and Unix dependency.
import warnings

# local
from x84.bbs.ansiwin import AnsiWindow, GLYPHSETS
from x84.bbs.session import getterminal, getch
from x84.bbs.output import echo

#: default command-key mapping.
PC_KEYSET = {'refresh': [unichr(12), ],
             'backspace': [unichr(8), unichr(127), ],
             'backword': [unichr(23), ],
             'enter': [u'\r', ],
             'exit': [unichr(27), ], }


class LineEditor(object):

    """
    This unicode line editor is unaware of its (y, x) position.

    It is great for prompting a quick phrase on any terminal,
    such as a ``login:`` prompt.
    """

    _hidden = False
    _width = 0
    # pylint: disable=R0913
    #         Too many arguments (7/5) (col 4)

    def __init__(self, width=None, content=u'', hidden=False,
                 colors=None, glyphs=None, keyset=None):
        """
        Class initializer.

        :param int width: the maximum input length.
        :param str content: given default content.
        :param str hidden: When non-False, a single 'mask' character for output.
        :param dict colors: optional dictionary containing key 'highlight'.
        :param dict glyphs: optional dictionary of window border characters.
        :param dict keyset: optional dictionary of line editing values.
        """
        self._term = getterminal()
        self.content = content or u''
        self.hidden = hidden
        self._width = width
        self._input_length = self._term.length(content)

        self._quit = False
        self._carriage_returned = False

        self.init_keystrokes(keyset=keyset or PC_KEYSET.copy())
        self.init_theme(colors=colors, glyphs=glyphs)

    def init_theme(self, colors=None, glyphs=None, hidden=False):
        """ Set color, bordering glyphs, and hidden attribute theme. """
        # set defaults,
        self.colors = {'highlight': self._term.reverse}
        self.glyphs = GLYPHSETS['thin'].copy()

        # allow user override
        if colors is not None:
            self.colors.update(colors)
        if glyphs is not None:
            self.glyphs.update(glyphs)
        if hidden:
            self.hidden = hidden

    def init_keystrokes(self, keyset):
        """ Sets keyboard keys for various editing keystrokes. """
        self.keyset = keyset
        self.keyset['refresh'].append(self._term.KEY_REFRESH)
        self.keyset['backspace'].append(self._term.KEY_BACKSPACE)
        self.keyset['backspace'].append(self._term.KEY_DELETE)
        self.keyset['enter'].append(self._term.KEY_ENTER)
        self.keyset['exit'].append(self._term.KEY_ESCAPE)

    @property
    def quit(self):
        """ Whether a 'quit' character has been handled, such as escape. """
        return self._quit

    @property
    def carriage_returned(self):
        """ Whether the carriage return character has been handled. """
        return self._carriage_returned

    @property
    def hidden(self):
        """
        When non-False, a single 'mask' character for hiding input.

        Used by password prompts.
        """
        return self._hidden

    @hidden.setter
    def hidden(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert value is False or 1 == len(value)
        self._hidden = value

    @property
    def width(self):
        """ Limit of characters to receive on input. """
        return self._width

    @width.setter
    def width(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._width = value

    def refresh(self):
        """
        Return string sequence suitable for refreshing editor.

        No movement or positional sequences are returned.
        """
        disp_lightbar = u''.join((
            self._term.normal,
            self.colors.get('highlight', u''),
            ' ' * self.width,
            '\b' * self.width))
        content = self.content
        if self.hidden:
            content = self.hidden * self._term.length(self.content)
        return u''.join((disp_lightbar, content, self._term.cursor_visible))

    def process_keystroke(self, keystroke):
        """
        Process the keystroke and return string to refresh.

        :param blessed.keyboard.Keystroke keystroke: input from ``inkey()``.
        :rtype: str
        :returns: string sequence suitable for refresh.
        """
        self._quit = False
        keystroke = hasattr(keystroke, 'code') and keystroke.code or keystroke
        if keystroke in self.keyset['refresh']:
            return u'\b' * self._term.length(self.content) + self.refresh()
        elif keystroke in self.keyset['backspace']:
            if len(self.content) != 0:
                len_toss = self._term.length(self.content[-1])
                self.content = self.content[:-1]
                return u''.join((
                    u'\b' * len_toss,
                    u' ' * len_toss,
                    u'\b',))
        elif keystroke in self.keyset['backword']:
            if len(self.content) != 0:
                ridx = self.content.rstrip().rfind(' ') + 1
                toss = self._term.length(self.content[ridx:])
                move = len(self.content[ridx:])
                self.content = self.content[:ridx]
                return u''.join((
                    u'\b' * toss,
                    u' ' * move,
                    u'\b' * move,))
        elif keystroke in self.keyset['enter']:
            self._carriage_returned = True
        elif keystroke in self.keyset['exit']:
            self._quit = True
        elif isinstance(keystroke, int):
            return u''
        elif (ord(keystroke) >= ord(' ') and
                (self._term.length(self.content) < self.width
                 or self.width is None)):
            self.content += keystroke
            return keystroke if not self.hidden else self.hidden
        return u''

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).

        Allows backspacing. Returns unicode text, or None when canceled.
        """
        self._carriage_returned = False
        self._quit = False
        echo(self.refresh())
        while not (self.quit or self.carriage_returned):
            inp = getch()
            echo(self.process_keystroke(inp))
        echo(self._term.normal)
        if not self.quit:
            return self.content
        return None


class ScrollingEditor(AnsiWindow):

    """
    A single line Editor, requires absolute (yloc, xloc) position.

    Infinite horizontal scrolling is enabled or limited using max_length.
    """

    # pylint: disable=R0902,R0904
    #        Too many instance attributes (14/7)
    #        Too many public methods (33/20)

    def __init__(self, *args, **kwargs):
        """
        Class initializer.

        :param int width: width of window.
        :param int yloc: y-location of window.
        :param int xloc: x-location of window.
        :param int max_length: maximum length of input (even when scrolled).
        :param dict colors: color theme.
        :param dict glyphs: bordering window character glyphs.
        :param dict keyset: command keys, global ``PC_KEYSET`` is default.
        """
        self._term = getterminal()
        self._horiz_shift = 0
        self._horiz_pos = 0
        # self._enable_scrolling = False
        self._horiz_lastshift = 0
        self._scroll_pct = kwargs.pop('scroll_pct', 25.0)
        self._margin_pct = kwargs.pop('margin_pct', 10.0)
        self._carriage_returned = False
        self._max_length = kwargs.pop('max_length', 0)
        self._quit = False
        self._bell = False
        self.content = kwargs.pop('content', u'')
        self._input_length = self._term.length(self.content)
        # there are some flaws about how a 'height' of a window must be
        # '3', even though we only want 1; we must also offset (y, x) by
        # 1 and width by 2: issue #161.
        kwargs['height'] = 3
        self.init_keystrokes(keyset=kwargs.pop('keyset', PC_KEYSET.copy()))
        AnsiWindow.__init__(self, *args, **kwargs)

    def init_theme(self, colors=None, glyphs=None):
        """ Set color and bordering glyphs theme. """
        AnsiWindow.init_theme(self, colors, glyphs)
        if 'highlight' not in self.colors:
            self.colors['highlight'] = self._term.yellow_reverse
        if 'strip' not in self.glyphs:
            self.glyphs['strip'] = u'$ '

    def init_keystrokes(self, keyset):
        """ Sets keyboard keys for various editing keystrokes. """
        self.keyset = keyset
        self.keyset['refresh'].append(self._term.KEY_REFRESH)
        self.keyset['backspace'].append(self._term.KEY_BACKSPACE)
        self.keyset['backspace'].append(self._term.KEY_DELETE)
        self.keyset['enter'].append(self._term.KEY_ENTER)
        self.keyset['exit'].append(self._term.KEY_ESCAPE)

    @property
    def position(self):
        """ Tuple of shift amount and column position of line editor. """
        return (self._horiz_shift, self._horiz_pos)

    @property
    def eol(self):
        """ Whether more input may be accepted (end of line reached). """
        return self._input_length >= self.max_length

    @property
    def bell(self):
        """ Whether the user has neared the margin. """
        margin = int(float(self.visible_width) *
                     (float(self.scroll_pct) * .01))
        return bool(self._input_length >= self.visible_width - margin)

    @bell.setter
    def bell(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._bell = value

    @property
    def carriage_returned(self):
        """ Whether the carriage return character has been handled. """
        return self._carriage_returned

    @property
    def quit(self):
        """ Whether a 'quit' character has been handled, such as escape. """
        return self._quit

    @property
    def is_scrolled(self):
        """ Whether the horizontal editor is in a scrolled state. """
        return bool(self._horiz_shift)

    @property
    def scroll_amt(self):
        """
        Number of columns from-end until horizontal editor will scroll

        Calculated by scroll_pct.
        """
        return int(float(self.visible_width) * (float(self.scroll_pct) * .01))

    @property
    def margin_amt(self):
        """
        Absolute number of columns from margin until bell is signaled.

        Indicating that the end is near and the carriage should be soon
        returned.
        """
        return int(float(self.visible_width) * (float(self.margin_pct) * .01))

    @property
    def scroll_pct(self):
        """
        Percentage of visible width from-end until scrolling occurs.

        Number of columns, as a percentage of its total visible width, that
        will be scrolled when a user reaches the margin by percent.  Default
        is 25.
        """
        return self._scroll_pct

    @scroll_pct.setter
    def scroll_pct(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._scroll_pct = float(value)
        assert value < 50, ("Bugs with values greater than 50 ...")

    @property
    def margin_pct(self):
        """
        Percentage of visible width from-end until bell is signaled.

        Number of columns away from input length limit, as a percentage of its
        total visible width, that will alarm the bell. This simulates the bell
        of a typewriter as a signaling mechanism. Default is 10.

        Unofficially intended for a faked multi-line editor: by using the bell
        as a wrap signal to instantiate another line editor and
        'return the carriage'.
        """
        return self._margin_pct

    @margin_pct.setter
    def margin_pct(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._margin_pct = float(value)

    @property
    def max_length(self):
        """
        Maximum line length.

        This also limits infinite scrolling when enable_scrolling is True.
        When unset, the maximum length is infinite!
        """
        if not self._max_length:
            warnings.warn("maximum length of ScrollingEditor is infinite!")
        return self._max_length or float('inf')

    @max_length.setter
    def max_length(self, value):
        # pylint: disable=C0111
        #        Missing docstring
        self._max_length = value

    @property
    def content(self):
        """ The contents of the editor. """
        return self._content

    @content.setter
    def content(self, value):
        # pylint: disable=C0111
        #        Missing docstring
        self._content = value
        self._input_length = self._term.length(value)

    def process_keystroke(self, keystroke):
        """
        Process the keystroke and return string to refresh.

        :param blessed.keyboard.Keystroke keystroke: input from ``inkey()``.
        :rtype: str
        :returns: string sequence suitable for refresh.
        """
        self._quit = False
        rstr = u''
        keystroke = hasattr(keystroke, 'code') and keystroke.code or keystroke
        if keystroke in self.keyset['refresh']:
            rstr = self.refresh()
        elif keystroke in self.keyset['backspace']:
            rstr = self.backspace()
        elif keystroke in self.keyset['backword']:
            rstr = self.backword()
        elif keystroke in self.keyset['enter']:
            self._carriage_returned = True
            rstr = u''
        elif keystroke in self.keyset['exit']:
            self._quit = True
            rstr = u''
        elif isinstance(keystroke, int):
            rstr = u''
        else:
            if ord(keystroke) >= 0x20:
                rstr = self.add(keystroke)
        return rstr

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).

        Allows backspacing. Returns unicode text, or None when canceled.
        """
        echo(self.refresh())
        self._quit = False
        self._carriage_returned = False
        while not (self.quit or self.carriage_returned):
            inp = getch()
            echo(self.process_keystroke(inp))
        if not self.quit:
            return self.content
        return None

    def fixate(self, x_adjust=0):
        """
        Return string sequence suitable for "fixating" cursor position.

        Set x_adjust to -1 to position cursor 'on' the last character,
        or 0 for 'after' (default).
        """
        xpos = self._xpadding + self._horiz_pos + x_adjust
        return self.pos(1, xpos) + self._term.cursor_visible

    def refresh(self):
        """
        Return string sequence suitable for refreshing editor.

        A strange by-product; if scrolling was not previously enabled, it is if
        wrapping must occur; this can happen if a non-scrolling editor was
        provided a very large .content buffer, then later .refresh()'d. --
        essentially enabling infinite scrolling.
        """
        # reset position and detect new position
        self._horiz_lastshift = self._horiz_shift
        self._horiz_shift = 0
        self._horiz_pos = 0
        #                  (self._term.length(self.content))
        for _ in range(self._input_length):
            if (self._horiz_pos >
                    (self.visible_width - self.scroll_amt)):
                self._horiz_shift += self.scroll_amt
                self._horiz_pos -= self.scroll_amt
            self._horiz_pos += 1
        if self._horiz_shift > 0:
            self._horiz_shift += len(self.glyphs['strip'])
            prnt = u''.join((
                self.glyphs['strip'],
                self.content[self._horiz_shift:],))
        else:
            prnt = self.content
        return u''.join((
            self.pos(self.ypadding, self.xpadding),
            self._term.normal,
            self.colors.get('highlight', u''),
            self.align(prnt),
            self.fixate(),))

    def backword(self):
        """
        Delete word behind cursor, using ' ' as boundary.

        In gnu-readline this is unix-word-rubout (C-w).
        """
        if 0 == len(self.content):
            return u''
        ridx = self.content.rstrip().rfind(' ') + 1
        self.content = self.content[:ridx]
        return self.refresh()

    def backspace(self):
        """ Remove character from end of buffer, scroll as necessary. """
        if 0 == len(self.content):
            return u''
        rstr = u''
        # measured backspace erases over double-wide (wcwidth)
        len_toss = self._term.length(self.content[-1])
        len_move = 1
        self.content = self.content[:-1]
        if self.is_scrolled and (self._horiz_pos < self.scroll_amt):
            # shift left,
            self._horiz_shift -= self.scroll_amt
            self._horiz_pos += self.scroll_amt
            rstr += self.refresh()
        else:
            rstr += u''.join((
                self.fixate(0),
                u'\b' * len_toss,
                u' ' * len_move,
                u'\b' * len_move,))
            self._horiz_pos -= 1
        return rstr

    def update(self, ucs=u''):
        """
        Replace or reset content.

        Resets properties ``carriage_returned`` and ``quit`` to False.
        """
        self._horiz_shift = 0
        self._horiz_pos = 0
        self.content = ucs
        self._carriage_returned = False
        self._quit = False
        assert unichr(27) not in ucs, ('Editor is not ESC sequence-safe')

    def add(self, u_chr):
        """
        Return output sequence of changes after adding a character to editor.

        An empty string is returned if no data could be inserted. Sequences for
        re-displaying the full input line are returned when the character
        addition caused the window to scroll horizontally.

        Otherwise, the input is simply returned to be displayed.
        """
        if self.eol:
            # cannot input, at end of line!
            return u''

        # append input to content directly to the backend variable,
        # so that we adjust the length only by the most recently-added
        # character.
        self._content += u_chr
        self._input_length += self._term.length(u_chr)

        # return character appended as output, ensure .fixate() is used first!
        self._horiz_pos += 1
        if self._horiz_pos >= (self.visible_width - self.margin_amt):
            # scrolling is required,
            return self.refresh()
        return self._term.normal + self.colors['highlight'] + u_chr



"""
User profile editor script for x/84.

This script is closely coupled with, and dependent upon nua.py.
"""
# std imports
from __future__ import division
import collections
import fnmatch
import string
import time

# bbs
from x84.bbs import getterminal, getsession, echo, timeago, get_ini, goto
from x84.bbs import list_users, find_user, get_user
from x84.bbs import LineEditor, ScrollingEditor

# local
from common import display_banner
import nua

#: lowlight color -- brackets ``[]``, and colon, ``:``
color_lowlight = get_ini(
    section='profile', key='color_lowlight'
) or 'bright_black'

#: highlight color -- read-only field values and hotkeys
color_highlight = get_ini(
    section='profile', key='color_highlight'
) or 'bright_magenta'

#: field display color -- read-only field values and hotkeys
color_field_display = get_ini(
    section='profile', key='color_field_display'
) or 'reverse_magenta'

#: field display color -- read-only field values and hotkeys
color_field_edit = get_ini(
    section='profile', key='color_field_edit'
) or 'reverse'

Point = collections.namedtuple('Point', ['y', 'x'])

field = collections.namedtuple('input_validation', [
    # field's value
    'value',
    # field format, u'{lb}{key}{rb}mail:    {email}'
    'field_fmt',
    # field text y/x loc,
    'display_location',
    # interactive edit field x/y loc, (None=read-only)
    'edit_location',
    # edit field key, (None=read-only)
    'key',
    # width of field (None=not justified)
    'width',
    # validating function
    'validate_fn',
    # text description for the content of this field,
    'description',
])


def get_display_fields(user, point):
    """ Return OrderedDict of display fields. """
    # reference the description text fields from nua.py
    # for their same purpose here.
    descriptions = {
        key: field.description for (key, field) in
        nua.get_validation_fields(user).items()}

    fields = collections.OrderedDict()
    _indent = point.x + 4 + nua.username_max_length
    # user: <name> last called 10m ago
    #              from ssh-127.0.0.1:65534
    #              1 calls, 1 posts
    fields['user'] = field(
        value=user.handle,
        field_fmt=u'{rb}{value}{lb}',
        display_location=Point(y=point.y, x=point.x),
        edit_location=Point(y=point.y, x=point.x + 1),
        key=None, width=nua.username_max_length, validate_fn=None,
        description=descriptions.get('handle')
    )
    fields['ago'] = field(
        value=timeago(time.time() - user.lastcall),
        field_fmt=u'last called {value} ago',
        display_location=Point(y=point.y, x=_indent),
        edit_location=Point(None, None),
        key=None, width=None, validate_fn=None, description=None,
    )
    fields['last_from'] = field(
        value=user.get('last_from', 'None'),
        field_fmt=u'from {value}',
        display_location=Point(y=point.y + 1, x=_indent),
        edit_location=Point(None, None),
        key=None, width=None, validate_fn=None, description=None,
    )
    fields['calls'] = field(
        value=str(user.calls),
        field_fmt=u'{value} calls',
        display_location=Point(y=point.y + 2, x=_indent),
        edit_location=Point(None, None),
        key=None, width=None, validate_fn=None, description=None,
    )
    fields['posts'] = field(
        value=str(user.get('msgs_sent', 0)),
        field_fmt=u'{value} posts',
        display_location=Point(y=point.y + 2,
                               x=_indent + len('1999 calls') + 1),
        edit_location=Point(None, None),
        key=None, width=None, validate_fn=None, description=None,
    )
    # go ahead, show them the password salt; it gets trimmed, and its gibberish,
    # maybe it gives them confidence that we don't know their actual password.
    _password = u''
    if user.handle != 'anonymous':
        _password = u''.join(user.password)
    fields['password'] = field(
        value=_password,
        field_fmt=u'{lb}{key}{rb}assword{colon} {value}',
        display_location=Point(y=point.y + 5, x=point.x),
        edit_location=Point(y=point.y + 5, x=point.x + 12),
        key=u'p', width=nua.password_max_length,
        validate_fn=nua.validate_password, description=None,
    )
    fields['location'] = field(
        value=user.location,
        field_fmt=u'{lb}{key}{rb}rigin{colon}   {value}',
        display_location=Point(y=point.y + 7, x=point.x),
        edit_location=Point(y=point.y + 7, x=point.x + 12),
        key=u'o', width=nua.location_max_length,
        validate_fn=None, description=descriptions.get('location'),
    )
    fields['email'] = field(
        value=user.email,
        field_fmt=u'{lb}{key}{rb}mail{colon}    {value}',
        display_location=Point(y=point.y + 9, x=point.x),
        edit_location=Point(y=point.y + 9, x=point.x + 12),
        key=u'e', width=nua.email_max_length,
        validate_fn=None, description=descriptions.get('email'),
    )
    fields['timeout'] = field(
        value=str(user.get('timeout', 'no')),
        field_fmt=u'{lb}{key}{rb}dle off{colon} {value}',
        display_location=Point(y=point.y + 11, x=point.x),
        edit_location=Point(y=point.y + 11, x=point.x + 12),
        key=u'i', width=5,
        # XXX
        validate_fn=None,
        description=(u"When set, Your session will be disconnected after this "
                     u"period of time has elapsed (in seconds).  0 disables."),
    )
    fields['pubkey'] = field(
        value=user.get('pubkey') or 'no',
        field_fmt=u'{lb}{key}{rb}sh-key{colon}  {value}',
        display_location=Point(y=point.y + 11, x=point.x + 19),
        edit_location=Point(y=point.y + 11, x=point.x + 31),
        key=u's', width=20,
        # XXX
        validate_fn=None,
        description=(u"Place your OpenSSH-compatible ssh public key into your "
                     u"clipboard, using command `pbcopy < ~/.ssh/id_rsa.pub` "
                     "on OSX or `xclip -i < ~/.ssh/id_rsa.pub` in X11.  Then, "
                     "simply paste it here. "),
    )
    fields['groups'] = field(
        value=','.join(user.groups),
        field_fmt=u'{lb}{key}{rb}roups{colon}   {value}',
        display_location=Point(y=point.y + 13, x=point.x),
        edit_location=Point(y=point.y + 13, x=point.x + 12),
        key=u'g', width=30,
        validate_fn=None,
        description=(u"Groups this user is a member of, separated by comma.  "
                     u"Notably, group 'sysop' has system-wide access, and "
                     u"group 'moderator' is able to moderate messages. "),
    )
    return fields


def show_banner(term):
    """ Display banner, calculate and return x/y coords as Point. """
    if term.height >= 24:
        echo(term.move(term.height - 1, 0))
        yloc = display_banner('art/ue.ans') + 1
    else:
        echo(term.move(term.height - 1, 0))
        # create a new, empty screen
        echo(u'\r\n' * (term.height + 1))
        yloc = 1
    return Point(y=yloc, x=max(5, (term.width // 2) - 30))


def display_options(term, fields):
    """ Display all fields with their values and edit-key commands. """
    _color1, _color2, _color3, _color4 = [
        getattr(term, _color)
        for _color in (color_lowlight,
                       color_highlight,
                       color_field_display,
                       color_field_edit)]

    lb, rb, colon = _color1('['), _color1(']'), _color1(':')
    out_text = u''
    for field_name, field in fields.items():
        # trim and padd field-value to maximum length,
        _align = term.center if field_name == 'user' else term.ljust
        _value = (
            # field value is greater than field width, trim to size
            u'.. ' + field.value[-(field.width - len(u'.. ')):]
            if field.width and term.length(field.value) > field.width
            # left-adjust string to length of field, except for user,
            # which is always centered
            else _align(field.value, field.width)
            if field.width
            # dynamic-length field, as-is
            else field.value)

        # backlight editable fields, bold others, except for 'user',
        # which has a special case of always being highlighted by
        # the 4th color of the palette.
        _value = (_color3(_value) if field.key
                  else _color4(_value) if field_name == 'user'
                  else _color2(_value))

        # bold keystroke if not None,
        _key = _color2(field.key) if field.key else None

        out_text += (
            u'{move_yx}{text}{clear_eol}'.format(
                move_yx=term.move(*field.display_location),
                text=field.field_fmt.format(
                    rb=rb, lb=lb, colon=colon, value=_value, key=_key),
                clear_eol=term.clear_eol))
    return out_text


def display_prompt(term, session, point):
    """ Display prompt. """
    # < [q]uit, [f]ind user, [d]elete > ?
    _color1, _color2, = [getattr(term, _color)
                         for _color in (color_lowlight, color_highlight)]
    lb, rb, colon = _color1('['), _color1(']'), _color1(':')

    out_text = term.move(*point)
    if session.user.is_sysop:
        out_text = u''.join(
            (out_text,
             u'{lb}{key_lt}{rb}prev, '
             .format(lb=lb, key_lt=_color2(u'<'), rb=rb)))

    out_text = u''.join(
        (out_text,
         u'{lb}{key_d}{rb}elete, '
         u'{lb}{key_q}{rb}uit'
         .format(lb=lb, rb=rb,
                 key_q=_color2(u'q'),
                 key_d=_color2(u'd'),
                 )))

    if session.user.is_sysop:
        # administrative functions (sysops only)
        out_text = u''.join(
            (out_text,
             u', {lb}{key_f}{rb}ind user, '
             u'{lb}{key_gt}{rb}next'
             .format(lb=lb, rb=rb,
                     key_f=_color2(u'f'),
                     key_gt=_color2(u'>'))))

    out_text = u''.join(
        (out_text,
         u'{colon}{clear_eos} ?\b\b'
         .format(colon=colon, clear_eos=term.clear_eos)))

    return out_text


def do_command(term, session, inp, fields, tgt_user, point):
    """ Perform action by given input. """
    _color1, _color2, _color3 = [
        getattr(term, _color) for _color in (
            color_lowlight, color_highlight, color_field_edit)]

    # discover edit field by matching command key
    field_name = None
    for _fname, field in fields.items():
        if field.key == inp.lower():
            field_name = _fname
            break

    if field_name is None:
        # return False if no field matches this key
        return False

    # only 'sysop' may edit user groups
    if field_name == 'groups' and not session.user.is_sysop:
        return False

    # pylint: disable=W0631
    #         Using possibly undefined loop variable 'field'

    # TODO: we could probably stand to do a bit of a better job of detecting
    # screen resizes and providing ^L full-screen refresh during the remainder
    # of this procedure ... It would require quite the refactor, though.

    # special case; ssh public keys and groups use scrolling editor
    if field_name in ('pubkey', 'groups'):
        editor = ScrollingEditor(
            # issue #161; because of a 'border' (that we don't draw),
            # y must be offset by 1 and height by 2.
            yloc=field.edit_location.y - 1,
            xloc=field.edit_location.x - 1,
            width=field.width + 2,
            colors={'highlight': _color3})
        # limit input to 1K
        editor.max_length = 1024
    else:
        editor = LineEditor(field.width, colors={'highlight': _color3})

    # find width for displaying description text and validation errors
    width = term.width - (point.x * 2)

    # show field description and cancellation
    description = (field.description or u'') + '  Press escape to cancel.'
    description_text = term.wrap(description, width=width)
    for y_offset, txt in enumerate(description_text):
        echo(term.move(point.y + y_offset, point.x))
        echo(_color1(txt) + term.clear_eol)
    echo(term.clear_eos)

    # edit input field (occludes display field).
    echo(term.move(*field.edit_location))
    inp = editor.read()
    if inp is None:
        # escape was pressed
        echo(term.move(*point))
        echo(_color2('Canceled !') + term.clear_eos)
        time.sleep(1)
        return True

    else:
        # validate input
        if field.validate_fn is not None:
            errmsg, _ = field.validate_fn(tgt_user, inp)
            if errmsg:
                # failed validation, display validation error
                errmsg += '  Press any key.'
                for y_offset, txt in enumerate(term.wrap(errmsg, width=width)):
                    echo(term.move(point.y + y_offset, point.x))
                    echo(_color2(txt) + term.clear_eol)
                echo(term.clear_eos)
                term.inkey()
                return True

        # well, it has validated, shall we apply it, then?
        if field_name in ('password', 'location', 'email',):
            # except for anonymous,
            if tgt_user.handle != 'anonymous':
                setattr(tgt_user, field_name, inp)
        elif field_name in ('timeout', 'pubkey',):
            if field_name == 'timeout':
                # coerce to integer, set, and if tgt_user is our current
                # user, then send new value for as engine event
                try:
                    timeout_val = int(inp)
                except ValueError:
                    return True
                if tgt_user.handle != 'anonymous':
                    tgt_user[field_name] = timeout_val
                if tgt_user.handle == session.user.handle:
                    session.send_event('set-timeout', timeout_val)
            elif field_name == 'pubkey':
                if tgt_user.handle != 'anonymous':
                    tgt_user[field_name] = inp
        elif field_name in ('groups'):
            new_groups = set(filter(None, set(map(unicode.strip, inp.split(',')))))
            for old_grp in tgt_user.groups.copy():
                if old_grp not in new_groups:
                    tgt_user.group_del(old_grp)
            for new_grp in new_groups:
                if new_grp not in tgt_user.groups:
                    tgt_user.group_add(new_grp)
        else:
            raise ValueError('unknown field name: {0}'.format(field_name))
    if tgt_user.handle != 'anonymous':
        tgt_user.save()
    return True


def delete_user(term, tgt_user, point):
    """ Delete given user. You may delete yourself. """
    _color1, _color2 = [getattr(term, _color)
                        for _color in (color_lowlight, color_highlight)]
    lb, rb, colon = _color1('['), _color1(']'), _color1(':')

    echo(term.move(*point))
    echo(u'Delete {handle} {lb}yN{rb}{colon}{clear_eos} ?\b\b'
         .format(handle=_color2(tgt_user.handle),
                 rb=rb, lb=lb, colon=colon,
                 clear_eos=term.clear_eos))
    inp = term.inkey()
    echo(inp + term.move(point.y + 2, point.x))
    if inp == u'y':
        if tgt_user.handle != 'anonymous':
            tgt_user.delete()
        echo(_color2('Deleted !'))
        time.sleep(1)
        return True

    echo(_color2('Canceled !'))
    time.sleep(1)
    return False


def locate_user(term, point):
    """ Prompt for search pattern and return discovered User. """
    _color1, _color2, _color3 = [
        getattr(term, _color) for _color in (
            color_lowlight, color_highlight, color_field_edit)]

    # show help
    width = term.width - (point.x * 2)
    help_txt = (u'Enter username or glob pattern.  Press escape to cancel.')
    y_offset = 0
    for y_offset, txt in enumerate(term.wrap(help_txt, width=width)):
        echo(term.move(point.y + y_offset, point.x))
        echo(_color1(txt) + term.clear_eol)
    point_prompt = Point(y=point.y + y_offset + 2, x=point.x)

    editor = LineEditor(nua.username_max_length, colors={'highlight': _color3})
    while True:
        # prompt for handle
        echo(term.move(*point_prompt))
        echo(u'handle: ' + term.clear_eol)
        inp = editor.read()

        point = Point(y=point_prompt.y + 2, x=point.x)
        if inp is None:
            # canceled (escape)
            return
        elif u'*' in inp or u'?' in inp:
            # a glob pattern, fetch all usernames
            handles = fnmatch.filter(list_users(), inp)
            if len(handles) == 0:
                echo(u''.join((term.move(*point),
                               u'No matches for {0}.'.format(_color2(inp)),
                               term.clear_eos)))
            elif len(handles) == 1:
                return get_user(handles[0])
            else:
                matches_text = (
                    u'{0} accounts matched, chose one: {1}.'.format(
                        _color2(str(len(handles))), u', '.join(
                            _color2(handle) for handle in handles)))
                echo(term.move(*point))
                for y_offset, txt in enumerate(
                        term.wrap(matches_text, width=width)):
                    echo(term.move(point.y + y_offset, point.x))
                    echo(txt + term.clear_eol)
                    if point.y + y_offset > term.height - 3:
                        # we simply cannot display anymore
                        break
                echo(term.clear_eos)
        else:
            handle = find_user(inp)
            if handle is not None:
                return get_user(handle)
            echo(u''.join((term.move(*point),
                           u'No matches for {0}.'.format(_color2(inp)),
                           term.clear_eos)))


def get_prev_user(tgt_user):
    """ Get previous user in sorted order. """
    handles = sorted(list_users())
    idx = max(handles.index(tgt_user.handle) - 1, 0)
    return get_user(handles[idx])


def get_next_user(tgt_user):
    """ Get next user in sorted order. """
    handles = sorted(list_users())
    try:
        current_idx = handles.index(tgt_user.handle)
    except ValueError:
        # what if we just deleted the target user?
        # inject it back into `handles' list and try again
        handles = sorted(handles + [tgt_user.handle])
        current_idx = handles.index(tgt_user.handle)
    idx = min(current_idx + 1, len(handles) - 1)
    try:
        return get_user(handles[idx])
    except KeyError:
        # and what if we deleted the last user of the list?
        # take the previous one.
        return get_user(handles[idx - 1])


def main(handle=None):
    """ Main procedure. """
    dirty = -1
    session, term = getsession(), getterminal()
    tgt_user = get_user(handle) if handle else session.user
    legal_input_characters = string.letters + u'<>'

    # re-display entire screen on loop,
    while True:

        # full-screen refresh on -1; otherwise only the fields (such as when an
        # edit has occurred).  dirty = -1 is set on screen resize, for example.
        if dirty == -1:
            # display banner, discover (y,x) point after,
            point_margin = show_banner(term)

            # forward-calculate the prompt (y,x) point
            point_prompt = Point(y=point_margin.y + 15, x=point_margin.x)

        # get all field values and locations,
        fields = get_display_fields(tgt_user, point=point_margin)

        # display all fields and prompt
        echo(display_options(term, fields))
        echo(display_prompt(term, session, point=point_prompt))

        dirty = 0

        # blocking loop until screen refresh or keystroke
        event = None
        while event is None:
            # This dual input/refresh trick only works here, receiving
            # raw (undecoded) keyboard data as 'inp' because we don't
            # require the use of any application keys or multi-byte
            # sequences, only alphabetic characters.
            event, data = session.read_events(('input', 'refresh'))
            if event == 'refresh':
                dirty = -1
                break

            inp = data
            if inp in legal_input_characters:
                # display command input
                echo(inp.decode('ascii'))

            if inp == u'q':
                # [q]uit
                echo(u'\r\n')
                return
            elif inp == u'\x0c':
                # [^L] refresh
                dirty = -1
                break
            elif inp == u'f' and session.user.is_sysop:
                tgt_user = locate_user(term, point_prompt) or tgt_user
                break
            elif inp == u'd':
                # yes, you can delete yourself !!
                if delete_user(term, tgt_user, point_prompt):
                    if tgt_user == session.user:
                        # but if you delete yourself,
                        # you must logoff.
                        goto('logoff')

                    # otherwise, move to next user
                    tgt_user = get_next_user(tgt_user)
                break
            elif inp == u'<' and session.user.is_sysop:
                tgt_user = get_prev_user(tgt_user)
                break
            elif inp == u'>' and session.user.is_sysop:
                tgt_user = get_next_user(tgt_user)
                break
            elif inp in string.letters:
                if do_command(
                        term, session, inp, fields, tgt_user, point_prompt):
                    # when returning True, perform full-screen refresh,
                    break
                else:
                    # otherwise, clean prompt field
                    time.sleep(0.2)
                    echo(u'\b \b')
            elif inp in legal_input_characters:
                # though legal, not authorized: clean prompt field
                time.sleep(0.2)
                echo(u'\b \b')
            event = None
        dirty = dirty or 1
